# ==============================================================================
# Modules
# ==============================================================================
# Built-ins
import os, sys, requests, time
# ======
# Future
# ======
# (1) Add ability for `path` arguments to be `pathlib.Path`
# (2) Add ability to process ~ in path
# =============
# Web
# =============
# Reading html from website
def read_url(url:str, params=None, **kwargs):
    """
    Future:
    Make wrapper for dynamic html and phantom js
    """
    dynamic = False
    if not dynamic:
        r = requests.get(url, params=params, **kwargs)
        return r.text
    else:
        print("Not finished. Need to make wrapper for dynamic HTML using PhantomJS", file=sys.stderr)



import os, sys, datetime, site, pathlib, operator
from collections import OrderedDict, defaultdict, namedtuple
import pandas as pd
import numpy as np
import ete3
import skbio
# import mmh3
from scipy import stats
from scipy.spatial.distance import squareform
from tqdm import tqdm
from ..symmetry import Symmetric
from ..io import write_object, read_fasta, read_textfile
from ..utils import pd_dataframe_matmul, format_path, pd_series_collapse, is_dict_like, is_query_class, is_symmetrical, fragment, assert_acceptable_arguments, hash_kmer
from ..transmute.normalization import normalize
from Bio.SeqIO.FastaIO import SimpleFastaParser

__all__ = ["PhylogenomicFunctionalComponents",  "phylogenomically_binned_functional_potential", "mcr_from_directory", "get_taxonomy_lineage_from_identifier", "get_taxonomy_identifier_from_name", "infer_taxonomy", "alpha_diversity", "beta_diversity", "otu_to_level", "reverse_complement","hash_kmer", "count_kmers", "prevalence", "get_intergenic_sequences"]
__all__ = sorted(__all__)

# Extract intergenic sequences
def get_intergenic_sequences(path_annotations:str, path_sequences:str, seq_type="gene", include_peripheral_info=True, peripheral_tag="locus_tag", into=pd.Series):
    """
    Currently tested on GFF3 but it should work for GTF too
    """
    # Get sequences
    sequences = read_fasta(path_sequences, description=False, into=dict, verbose=False)

    # Get gene positions
    d_contig_gene_info = defaultdict(list)

    for i, line in read_textfile(path_annotations):
        if bool(line):
            fields = line.split("\t")
            id_contig = fields[0]
            if fields[2] == "gene":
                pos_start = int(fields[3])-1
                pos_end = int(fields[4])
                data = dict(map(lambda item:item.split("="), filter(bool, fields[-1].split(";"))))
                d_contig_gene_info[id_contig].append((pos_start, pos_end, fields[6], data[peripheral_tag]))

    # Get intergenic sequences
    d_id_seqs= OrderedDict()
    for id_contig, info in d_contig_gene_info.items():
        seq_reference = sequences[id_contig]
        for i in range(0, len(info)-1):
            pos_start = info[i][1]
            pos_end = info[i+1][0]
            id = "{}|intergenic_{}-{}".format(id_contig,pos_start+1, pos_end)
            if include_peripheral_info:
                id = "{}|({}){}|({}){}".format(id,info[i][2], info[i][3], info[i+1][2], info[i+1][3])
            seq_intergenic = seq_reference[pos_start:pos_end]
            if len(seq_intergenic) > 0:
                d_id_seqs[id] = seq_intergenic
    return into(d_id_seqs)


# Phylogeneomically binned functional potential
def phylogenomically_binned_functional_potential(X:pd.DataFrame,MCR:pd.DataFrame):
    """
    X: DataFrame of scaled compositional data that preserves proportionality (e.g. TPM or relative-abundance)
    MCR: DataFrame

    X.shape (n,m)
    MCR.shape (m,p)
    Output: (n,p)
    """
    assert X.shape[1] == MCR.shape[0], "Shapes are not compatible"
    return pd_dataframe_matmul(X,MCR)

# Parse the files from KEGG Maple
def mcr_from_directory(directory:str, name=None, end_tag="module.xls", mcr_field="MCR % (ITR)", index_name_bins = "id_bin", index_name_module = "id_module", store_raw_data=False):
    """
    directory:
    Path to kegg maple results which should have the following structure:
    | directory
    | | bin_1
    | | | xyz1.module.xls
    | | bin_2
    | | | xyz2.module.xls

    Returns:
    namedtuple with the following attributes:
        .ratios = MCR ratio
        .metadata = module metadata
        .name = user-provided name of experiment
        .raw_data = {id_bin:raw_kegg_maple_}

    Assumes subdirectories are bin_ids and
    each directory has a file that has the
    following format: 8206709269.module.xls

    https://www.genome.jp/maple-bin/mapleSubmit.cgi?aType=sDirect
    """
    MCR = namedtuple("MCR", ["ratios", "metadata", "name", "raw_data"])
    # Format the path
    directory = format_path(directory)
    # Convert the path to a Path object
    directory = pathlib.Path(directory)

    # Check directory
    assert directory.exists(), f"{str(directory.absolute())} does not exist"
    assert directory.is_dir(), f"{str(directory.absolute())} is not a directory"

    # Iterate through each KEGG Maple result dataset
    d_bin_mcr = OrderedDict()
    module_metadata = list()
    d_bin_raw = dict()
    for subdir in tqdm(directory.iterdir(), f"Reading subdirectories in {str(directory)}"):
        sheets = list()
        if subdir.is_dir():
            id_bin = subdir.name
            for file in subdir.iterdir():
                if file.name.endswith(end_tag):
                    for sheetname in ["module_pathway", "module_complex", "module_function", "module_signature"]:
                        df = pd.read_excel(file, index_col=0, sheetname=sheetname).set_index("ID", drop=False)
                        sheets.append(df)
            df_bin = pd.concat(sheets, axis=0)
            module_metadata.append(df_bin.loc[:,["Type", "Small category", "Name(abbreviation)"]])
            d_bin_mcr[id_bin] = df_bin[mcr_field]
            # Store raw data
            if store_raw_data:
                d_bin_raw[id_bin] = df_bin.drop("ID", axis=1)
    # Module Completion Ratios
    df_mcr = pd.DataFrame(d_bin_mcr).T.fillna(0.0)
    df_mcr.index.name = index_name_bins
    df_mcr.columns.name = index_name_module

    # Metadata
    df_metadata = pd.concat(module_metadata, axis=0).drop_duplicates() # May not be the most efficient but w/e

    return MCR(ratios=df_mcr, metadata=df_metadata, name=name, raw_data=d_bin_raw)


# Phylogenomic Functional Components
class PhylogenomicFunctionalComponents(object):
    __doc__ = """
    PhylogenomicFunctionalComponents
    ================================
    This includes methods to create Phylogenomic Functional Components.
    For `components_taxonomy` check out the databases available with Soothsayer:
    %s/soothsayer/db/

    Formatting should resemble the following:

    components_taxonomy.head()
    NODE_100000_length_1286_cov_38.26_32_868_-            UNBINNED__Streptococcus mutans
    NODE_100000_length_1286_cov_38.26_996_1285_-           UNBINNED__Streptococcus mitis
    NODE_100001_length_1286_cov_31.4249_1_152_+               UNBINNED__Neisseria mucosa
    NODE_100001_length_1286_cov_31.4249_257_925_+             UNBINNED__Neisseria mucosa
    NODE_100004_length_1286_cov_2.62551_1_686_-      UNBINNED__Leptotrichia goodfellowii
    Name: taxonomy, dtype: object

    components_functional.head() # Note this one will most likely have multiple functions for each ORF which is why it's a set and not a string
    NODE_100009_length_1286_cov_12.6458_1_870_+         {M00432__Leucine biosynthesis, 2-oxoisovalerate => 2-oxoisocaproate}
    NODE_395196_length_426_cov_39.5364_1_426_+       {M00051__Uridine monophosphate biosynthesis, glutamine (+ PRPP) => UMP}
    NODE_395187_length_426_cov_2.6442_1_426_-                         {M00120__Coenzyme A biosynthesis, pantothenate => CoA}
    NODE_39517_length_2534_cov_7.92457_605_1108_+                                             {M00063__CMP-KDO biosynthesis}
    NODE_39517_length_2534_cov_7.92457_1_572_+        {M00018__Threonine biosynthesis, aspartate => homoserine => threonine}
    """%site.getsitepackages()[0]
    def __init__(self,
                 # Input data
                 lengths:pd.Series,
                 components_taxonomy:pd.Series,
                 components_functional:pd.Series,
                 # Labels
                 label_unknown:str="unknown_function_collective",
                 attr_type="ORF",
                 obsv_type=None,
                 functional_type=None,
                 name=None,
                 description=None,
                 # Misc
                 delimiter = "|",
                ):
        # Input data
        self.lengths = lengths
        self.components_taxonomy = components_taxonomy
        self.components_functional = components_functional
        # Description
        self.name = name
        self.description = description
        # Data Types
        self.attr_type = attr_type
        self.obsv_type = obsv_type
        self.functional_type = functional_type
        # Miscellaneous
        self.label_unknown = label_unknown
        self.delimiter = delimiter
        self.__synthesized__ = datetime.datetime.utcnow()
        # ==========
        # Initialize
        # ==========
        print(f"Phylogenomic Functional Components",
               "==================================",
               sep="\n",
               file=sys.stderr,
         )
        # Assign ORFs to PGFCs
        self.pgfc_collection = self._assign_orfs_to_pgfcs()
        # Compile metadata for each PGFC
        self.metadata = self._get_metadata()
        # Datasets
        self.datasets = OrderedDict()
    def __repr__(self):
        class_name = str(self.__class__).split(".")[-1][:-2]
        return f"{class_name}(name:{self.name}, num_datasets:{len(self.datasets)}, num_pgfcs:{self.metadata.shape[0]})"

    # Assign ORFS to each PGFC
    def _assign_orfs_to_pgfcs(self):
        pgfc_collection = defaultdict(list)
        for taxonomy_component, Se_data in tqdm(self.components_taxonomy.groupby(self.components_taxonomy), f"Assigning ORFs"):
            idx_orfs_for_taxonomycomponent = set(Se_data.index)
            # Process missing functions
            idx_orfs_with_function = set(self.components_functional.index)
            idx_orfs_no_function = idx_orfs_for_taxonomycomponent - idx_orfs_with_function
            if len(idx_orfs_no_function) > 0:
                id_pgfc = self.delimiter.join([taxonomy_component, self.label_unknown])
                pgfc_collection[id_pgfc] = list(idx_orfs_no_function)
                del id_pgfc # Reset the namespace
            # Process those with functional category
            for id_orf, functional_list in self.components_functional[idx_orfs_for_taxonomycomponent & idx_orfs_with_function].iteritems():
                for functional_component in functional_list:
                    id_pgfc = self.delimiter.join([taxonomy_component, functional_component])
                    pgfc_collection[id_pgfc].append(id_orf)
        # Convert list to non-redundant sets
        return {id_pgfc:set(orfs) for id_pgfc, orfs in pgfc_collection.items()}

    # Get metadata for PGFCs
    def _get_metadata(self):
        metadata = defaultdict(dict)
        for id_pgfc, orfs in tqdm(self.pgfc_collection.items(), "Compiling metadata"):
            # Length
            query_lengths = self.lengths[orfs].values
            try:
                k2, p = stats.normaltest(query_lengths)
            except ValueError:
                k2 = p = np.nan
            metadata[id_pgfc]["orfs"] = orfs
            metadata[id_pgfc]["number_of_orfs"] = len(query_lengths)
            metadata[id_pgfc]["lengths"] = list(query_lengths)
            metadata[id_pgfc]["µ(lengths)"] = query_lengths.mean()
            metadata[id_pgfc]["∑(lengths)"] = query_lengths.sum()
            metadata[id_pgfc]["sem(lengths)"] = stats.sem(query_lengths)
            metadata[id_pgfc]["var(lengths)"] = np.var(query_lengths)
            metadata[id_pgfc]["normaltest__statistic"] = k2
            metadata[id_pgfc]["normaltest__pvalue"] = p
            metadata[id_pgfc]["taxonomy_component"] = id_pgfc.split(self.delimiter)[0]
            if id_pgfc.endswith(self.label_unknown):
                metadata[id_pgfc]["functional_component"] = np.nan
            else:
                metadata[id_pgfc]["functional_component"] = id_pgfc.split(self.delimiter)[-1]
        # Metadata
        df_meta = pd.DataFrame(metadata).T.loc[:,["taxonomy_component", "functional_component","number_of_orfs", "∑(lengths)", "µ(lengths)","sem(lengths)","var(lengths)","normaltest__statistic", "normaltest__pvalue", "lengths", "orfs"]].sort_values(["taxonomy_component", "functional_component", "number_of_orfs"], ascending=[True, True, False])
        df_meta.index.name = "id_phylogenomic-functional-component"

        # Ints
        for id_field in ["number_of_orfs", "∑(lengths)"]:
            df_meta[id_field] = df_meta[id_field].astype(int)
        # Floats
        for id_field in ["µ(lengths)","sem(lengths)","var(lengths)","normaltest__statistic", "normaltest__pvalue"]:
            df_meta[id_field] = df_meta[id_field].astype(float)
        return df_meta

    # Compute the summed ORF counts for each PGFC
    def compute_dataset(self, df_orfs:pd.DataFrame, name=None, copy_data=True, return_dataset=True, **attr):
        assert all(set(self.components_taxonomy.index) <= query_gene_set for query_gene_set in [set(self.lengths.index), set(df_orfs.columns)]), f"All genes in `components_taxonomy` must be in `lengths` and `df_orfs`"
        assert set(df_orfs.columns) <= set(self.lengths.index), f"All genes in `df_orfs` must be in `lengths`"
        components_taxonomy = self.components_taxonomy.dropna()
        components_functional = self.components_functional.dropna()

        # Name
        if name is None:
            name = len(self.datasets)

        # Sum ORFs for each PGFC
        d_pgfc_counts = dict()
        for id_pgfc, orfs in tqdm(self.pgfc_collection.items(), f"Summing ORFs for dataset = `{name}`"):
            d_pgfc_counts[id_pgfc] = df_orfs.loc[:,orfs].sum(axis=1) # Removed set construction here because orfs are non-redundant from _assign_orfs_to_pgfcs
        df_pgfc = pd.DataFrame(d_pgfc_counts).sort_index(axis=1)

        # Store dataset
        if copy_data:

            self.datasets[name] = {
                "shape_input":df_orfs.shape,
                "counts":df_pgfc,
                "shape_output":df_pgfc.shape,
                **attr,
            }
        if return_dataset:
            return df_pgfc
        else:
            return self
    # Get PGFC counts
    def get_dataset(self, name):
        assert name in self.datasets, f"{name} not in `datasets`.  Try`compute_dataset` and repeat."
        return self.datasets[name]["counts"]
    # Write object to file
    def to_file(self, path:str, compression="infer"):
        write_object(self, path=path, compression=compression)
        return self

# Get taxonomy lineage
def get_taxonomy_lineage_from_identifier(identifiers, name=None, translate_ids=True, ncbi:ete3.NCBITaxa=None, verbose=0, mode="infer", include_taxid=True, include_levels=["id_taxon", "phylum", "class", "order", "family", "genus", "species"]):
    """
    Input: Single Taxonomy ID or a collection of identifiers w/ {id_orf:id_taxon}
    Output: pd.Series/pd.DataFrame of rank and taxonomy labels
    modes: {'infer', 'batch', 'singular'}
    """
    accepted_modes = {'batch', 'singular'}
    verbose = int(verbose)
    # Get database
    if ncbi is None:
        ncbi = ete3.NCBITaxa()

    # Infer mode
    if mode == "infer":
        if is_dict_like(identifiers):
            mode = "batch"
        else:
            mode = "singular"
        if verbose > 0:
            print(f"Inferred mode: {mode}", file=sys.stderr)
        assert mode != "infer", "Cannot infer `mode`.  Please explicitly provide mode."

    # Singular
    if mode == "singular":
        # Handle missing data
        if pd.isnull(identifiers):
            return pd.Series([], name=name)

        # Access the database
        try:
            id_taxon = int(identifiers)
            if name is None:
                name = id_taxon
            lineage = ncbi.get_lineage(id_taxon)
            ranks = dict(filter(lambda x: x[1] != "no rank", ncbi.get_rank(lineage).items()))
            Se_taxonomy = pd.Series(ncbi.get_taxid_translator(ranks.keys()), name=name)
            if translate_ids:
                Se_taxonomy.index = Se_taxonomy.index.map(lambda x:ranks[x])
            return Se_taxonomy

        # Handle taxonomy IDs that are not in database
        except ValueError:
            if verbose > 1:
                print(id_taxon, file=sys.stderr)
            return pd.Series([], name=name)
    # Batch
    if mode == "batch":
        if not is_query_class(identifiers, "Series"):
            identifiers = pd.Series(identifiers)

        # Group each taxonomy identifier
        dataframes = list()
        for id_taxon, group in tqdm(pd_series_collapse(identifiers).iteritems(), "Searching lineage from taxonomy identifier"):
            number_of_orfs_in_group = len(group)
            Se_taxonomy = get_taxonomy_lineage_from_identifier(identifiers=id_taxon, name=None, translate_ids=translate_ids, ncbi=ncbi, verbose=verbose, mode="singular")
            df_taxon = pd.DataFrame(number_of_orfs_in_group*[Se_taxonomy])
            df_taxon.index = group
            if include_taxid:
                df_taxon["id_taxon"] = id_taxon
            dataframes.append(df_taxon)
        df_collection = pd.concat(dataframes, axis=0)
        if include_levels is None:
            include_levels = df_collection.columns
        else:
            include_levels = [*filter(lambda level:level in df_collection.columns, include_levels)]

        idx_missing_orfs = set(identifiers.index) - set(df_collection.index)
        number_of_orfs_missing = len(idx_missing_orfs)
        if number_of_orfs_missing > 0:
            A = np.empty((number_of_orfs_missing, df_collection.shape[1]))
            A[:] = np.nan
            df_missing = pd.DataFrame(A, index=idx_missing_orfs, columns=df_collection.columns)
            return pd.concat([df_collection, df_missing], axis=0).loc[identifiers.index, include_levels]
        else:
            return df_collection.loc[identifiers.index, include_levels]

# Get taxonomy identifiers from string name_taxonomy
def get_taxonomy_identifier_from_name(name_taxonomy, mode="infer", ncbi=None, into=pd.Series):
    """
    name_taxonomy can be either a single string or iterable of strings
    """
    accepted_modes = {'batch', 'singular'}

    # Get database
    if ncbi is None:
        ncbi = ete3.NCBITaxa()

    if mode == "infer":
        if not is_nonstring_iterable(name_taxonomy):
            mode = "singular"
        else:
            mode = "batch"
        assert mode != "infer", "Cannot infer `mode`.  Please explicitly provide mode."

    if mode == "singular":
        name_taxonomy = [name_taxonomy]
    name_taxonomy = [*map(lambda x:x.strip(),name_taxonomy)]
    Se_taxonomy = pd.Series(ncbi.get_name_translator(name_taxonomy)).sort_index().map(lambda x:x[0])
    if mode == "singular":
        return Se_taxonomy.values[0]
    if mode == "batch":
        idx_missing = set(name_taxonomy) - set(Se_taxonomy.index)
        number_of_name_taxonomy_missing = len(idx_missing)
        if number_of_name_taxonomy_missing > 0:
            Se_missing = pd.Series([np.nan]*number_of_name_taxonomy_missing, index=idx_missing)
            Se_taxonomy = pd.concat([Se_taxonomy, Se_missing])
        return into(Se_taxonomy)

# Infer taxonomy for a sequence grouping
def infer_taxonomy(
        grouping_sequence:pd.Series,
        grouping_taxonomy:pd.Series,
        percent_identities:pd.Series,
        grouping_protein_domains:pd.Series=None,
        include_protein_domains=None
    ):
    """
    Format:
    group_sequence:pd.Series where {id_orf:contig-or-bin-identifier}
    grouping_taxonomy:pd.Series where {id_orf:taxonomy-label}
    percent_identities:pd.Series where {id_orf:percent-identity}
    grouping_protein_domains:pd.Series where {id_orf:[PFAM_1, TIGRFAM_2, etc.]} (Optional)
    include_protein_domains: iterable of target protein-domain names (Optional)
    """
    # Comparable set
    idx_comparable = grouping_sequence.index & grouping_taxonomy.index & percent_identities.index

    # Subset data based on HMMs
    if include_protein_domains is not None:
        assert grouping_protein_domains is not None, "If `include_protein_domains` is provided then you must also provide `grouping_protein_domains`"
        include_protein_domains = set(include_protein_domains)
        grouping_protein_domains = grouping_protein_domains[grouping_protein_domains.map(bool)]
        mask_include_protein_domains = grouping_protein_domains.map(lambda x: bool(set(x) & include_protein_domains))
        grouping_protein_domains = grouping_protein_domains[mask_include_protein_domains]
        idx_comparable = idx_comparable & grouping_protein_domains.index

    # Convert percent identities to ratios
    if percent_identities.max() > 1.0:
        percent_identities = percent_identities/100

    # Infer taxonomy
    d_seqgroup_hits = defaultdict(lambda: defaultdict(int))
    d_seqgroup_weights = defaultdict(lambda: defaultdict(float))

    data = pd.DataFrame([grouping_sequence, grouping_taxonomy, percent_identities]).T.loc[idx_comparable,:].values

    for i, (id_orf, row_data) in tqdm(enumerate(zip(idx_comparable, data)), "Summing weights from ORFs"):
        id_seqgroup, id_taxon, p_identity = row_data
        d_seqgroup_hits[id_seqgroup][id_taxon] += 1
        d_seqgroup_weights[id_seqgroup][id_taxon] += p_identity

    # Calculate scores
    d_seqgroup_inferred = defaultdict(dict)
    for id_seqgroup, d_taxa_weight in tqdm(d_seqgroup_weights.items(), "Scoring taxonomy for each sequence group"):
        w = pd.Series(d_taxa_weight, name=id_seqgroup)
        scores = w/w.sum()
        d_seqgroup_inferred[id_seqgroup]["weights"] = sorted(w.iteritems(), key=lambda x:x[-1], reverse=True)
        max_score = scores.max()
        d_seqgroup_inferred[id_seqgroup]["score"] = max_score
        if max_score < 1:
            max_scoring_taxa = scores[lambda x: x == max_score].index
            if max_scoring_taxa.size > 1:
                inferred_taxonomy = list(max_scoring_taxa) # Added 2018-11-30
            else:
                inferred_taxonomy = max_scoring_taxa[0]
        else:
            inferred_taxonomy = scores.index[0]
        d_seqgroup_inferred[id_seqgroup]["inferred_taxonomy"] = inferred_taxonomy
        d_seqgroup_inferred[id_seqgroup]["number_of_total_hits"] = sum(d_seqgroup_hits[id_seqgroup].values())

    df_inferred = pd.DataFrame(d_seqgroup_inferred).T.loc[:,[ "inferred_taxonomy", "score", "number_of_total_hits", "weights"]]
    df_inferred.index.name = "id_seqgroup"
    return df_inferred

# Group Otus by higher taxonomy level
def otu_to_level(X:pd.DataFrame, taxonomy:pd.DataFrame, level):
    """
    X => pd.DataFrame of Otu counts with columns as Otus and rows as samples
    taxonomy => pd.DataFrame of taxonomy assignments for Otus
    level => a taxonomy level in the `taxonomy` dataframe
    """
    return X.groupby(taxonomy[level], axis=1).sum()

# Alpha diversity calculations
def alpha_diversity(X:pd.DataFrame, metric="richness", taxonomy:pd.DataFrame=None, mode="infer", idx_taxonomy=['Phylum', 'Class', 'Order', 'Family', 'Genus', 'Species'],  name=None, base=2, obsv_type=None, **alpha_kws):
    """
    X => pd.DataFrame of Otu counts with columns as Otus and rows as samples
    metric => a callable or a string {entropy, richness, gini, singletons}
    taxonomy => pd.DataFrame of taxonomy assignments for Otus
    """
    # Alpha measures
    def _entropy(x, **alpha_kws):
        return stats.entropy(x, base=base, **alpha_kws)
    def _richness(x, **alpha_kws):
        return (x > 0).sum().astype(int)
    def _gini(x, **alpha_kws):
        return skbio.diversity.alpha.gini_index(x, **alpha_kws)
    def _singletons(x, **alpha_kws):
        return (x == 1).sum().astype(int)
    d_metric_fn = {"entropy":_entropy, "richness":_richness, "gini":_gini, "singletons":_singletons}

    # Supported diversity measures
    if hasattr(metric, "__call__"):
        func = metric
    else:
        supported_metrics =  list(d_metric_fn.keys())
        assert metric in supported_metrics, f"`{metric}` is not compatible.  Only available alpha diversity measures are {supported_metrics}"
        func =  d_metric_fn[metric]
        name = metric

    # Compute diversity
    if mode == "infer":
        mode = "batch" if taxonomy is not None else "singular"

    assert mode in {"singular", "batch"}, "Please specify either 'singular', 'batch', or 'infer' for the mode"
    if mode == "singular":
        Se_alpha = X.apply(lambda x: func(x, **alpha_kws), axis=1)
        Se_alpha.index.name = name
        return Se_alpha

    if mode == "batch":
        assert taxonomy is not None, "`taxonomy` cannot be `None` when `mode='batch'`"
        d_level_metric = OrderedDict()
        for level in idx_taxonomy:
            if level in taxonomy.columns:
                df_level = otu_to_level(X, taxonomy, level=level)
                d_level_metric[level] = alpha_diversity(df_level, metric=metric, taxonomy=None, mode="singular", base=base, **alpha_kws)
            else:
                print(f"Skipping taxonomy level `{level}` because it is not in the taxonomy dataframe", file=sys.stderr)

        d_level_metric["Otu"] = alpha_diversity(X, metric=metric, taxonomy=None, mode="singular", base=base, **alpha_kws)
        df_level_metric = pd.DataFrame(d_level_metric)
        df_level_metric.index.name = f"id_{obsv_type}"
        return df_level_metric


# Compute the beta diversity
def beta_diversity(X:pd.DataFrame, y:pd.Series, label_intra="Intra", label_inter="Inter", label_metric="Distance", class_type="Class"):
    """
    X: a symmetric pd.DataFrame distance matrix with a diagonal of 0
    y: a pd.Series of class labels
    """
    if isinstance(X, Symmetric):
        X = X.to_dense()
    # Assertions
    assert is_symmetrical(X, tol=1e-10), "X must be symmetric"
    assert np.all(X.index == X.columns), "X.index != X.columns"
    assert set(X.index) <= set(y.index), "Not all elements in X.index are in y.index"
    assert np.all(X.notnull()), "X cannot contain any NaN.  Please preprocess the data accordingly"

    # Get indexable arrays
    idx_labels = X.index
    loc_labels = np.arange(X.shape[0])
    A = X.values

    # Calculate the diversity
    diversity_data = defaultdict(dict)
    for id_class, idx_class in pd_series_collapse(y[idx_labels], type_collection=pd.Index).items():
        loc_class = idx_class.map(lambda x:idx_labels.get_loc(x)).values
        loc_nonclass = np.delete(loc_labels, loc_class)
        # Intra
        intra_distances = squareform(A[loc_class,:][:,loc_class])
        diversity_data[(id_class, label_intra)] = pd.Series(intra_distances)
        # Inter
        inter_distances = A[loc_class,:][:,loc_nonclass].ravel()
        diversity_data[(id_class, label_inter)] = pd.Series(inter_distances)

    # Create output
    df_beta_diversity = pd.DataFrame(diversity_data)
    df_beta_diversity.columns.names = [class_type, "Diversity"]
    return df_beta_diversity





# Reverse Complement
def reverse_complement(seq:str):
    """
    Inputs a sequence of DNA and returns the reverse complement se$
    Outputs a reversed string of the input where A>T, T>A, G>C, an$
    """
    conversion = str.maketrans('ACGTacgt','TGCAtgca')
    comp = seq.translate(conversion)
    rev_comp = comp[::-1]
    return rev_comp

# # Murmurhash of K-mer
# def hash_kmer(kmer, random_state=0):
#     """
#     Adapted from the following source:
#     https://sourmash.readthedocs.io/en/latest/kmers-and-minhash.html
#     """
#     kmer = kmer.upper()

#     # Calculate the reverse complement
#     r_kmer = reverse_complement(kmer)

#     # Determine whether original k-mer or reverse complement is lesser
#     if kmer < r_kmer:
#         canonical_kmer = kmer
#     else:
#         canonical_kmer = r_kmer

#     # Calculate murmurhash using a hash seed
#     hash = mmh3.hash(canonical_kmer, seed=random_state, signed=False)
# #     if hash < 0:
# #         hash += 2**64

    return hash

def count_kmers(sequences, K=4, frequency=True, collapse_reverse_complement=False, pseudocount=0, fill_missing_kmers=False):
    """
    sequences: (1) a sequence string, (2) a dict_like of (id,seq) pairs, or (3) path to fasta
    collapse_reverse_complement reduces data dimensionality
    """
    def _infer_mode(sequences):
        mode = None
        if isinstance(sequences, str):
            if not os.path.exists(sequences):
                return "string_sequence"
            if os.path.exists(sequences):
                return "filepath"
        if is_dict_like(sequences):
            return "dict_like"
        assert mode is not None, "Unrecognized format:  Please provide: (1) a sequence string, (2) a dict_like of (id,seq) pairs, or (3) path to fasta"


    # Get mapping to collapse a strand
    def _collapse(kmers):
        kmer_collapse = dict()
        for kmer in kmers:
            r_kmer = reverse_complement(kmer)
            if r_kmer in kmer_collapse:
                kmer_collapse[kmer] = r_kmer
            else:
                kmer_collapse[kmer] = kmer
        return kmer_collapse

    def _count(seq:str, K, frequency, collapse_reverse_complement,  pseudocount, name=None):
        """
        Collapse only works for nondegenerate kmers
        """
        seq = seq.upper()
        alphabet = {"A", "C", "G", "T"}
        #query_alphabet = set(list(seq))
        # degenerate_kmers = {'B', 'D', 'H', 'K', 'M', 'N', 'R', 'S', 'V', 'W', 'Y'}
        #assert query_alphabet <= alphabet, "This cannot be used with degenerate characters"

        # Count kmers
        kmer_counts = defaultdict(int)
        for kmer in fragment(seq=seq, K=K, step=1, overlap=True):
            if set(kmer) <= alphabet:
                kmer_counts[kmer] += 1
        kmer_counts = pd.Series(kmer_counts, name=name) + pseudocount

        # Collapse double stranded
        if collapse_reverse_complement:
            kmer_collapse = _collapse(sorted(kmer_counts.index))
            kmer_counts = kmer_counts.groupby(kmer_collapse).sum()

        # Normalize
        kmer_counts = kmer_counts + pseudocount
        if frequency:
            return kmer_counts/kmer_counts.sum()
        if not frequency:
            return kmer_counts

    # Infer mode
    mode = _infer_mode(sequences)

    # Single sequence (string)
    if mode == "string_sequence":
        return _count(seq, K=K, frequency=frequency, collapse_reverse_complement=collapse_reverse_complement, pseudocount=pseudocount)
    else:
        # Multiple sequences
        params = dict(K=K, frequency=False, collapse_reverse_complement=False, pseudocount=0)
        f_sequences = None
        # Dict-like
        if mode == "dict_like":
            iterable = sequences.items()

        # Read in files
        if mode == "filepath":
            f_sequences = open(sequences, "r")
            iterable = SimpleFastaParser(f_sequences)

        # Count kmers
        data = list()
        for id, seq in tqdm(iterable, "Counting k-mers"):
            data.append(_count(name=id, seq=seq, **params))
        # Close file handle if necessary
        if f_sequences is not None:
            f_sequences.close()
        df_kmer_counts = pd.DataFrame(data)

        # Fill empty
        if fill_missing_kmers:
            missing_kmers = set(["".join(kmer) for kmer in itertools.product("ATCG",repeat=K)]) - set(df_kmer_counts.columns)
            A = np.zeros((df_kmer_counts.shape[0], len(missing_kmers)))
            df_missing = pd.DataFrame(A, index=df_kmer_counts.index, columns=list(missing_kmers))
            df_kmer_counts = pd.concat([df_kmer_counts, df_missing], axis=1)

        # Sort K-mers and fill null
        df_kmer_counts = df_kmer_counts.sort_index(axis=1, ascending=True).fillna(0)

        # Collapse Double-Stranded
        if collapse_reverse_complement:
            kmer_collapse = _collapse(df_kmer_counts.columns)
            df_kmer_counts = df_kmer_counts.groupby(kmer_collapse, axis=1).sum()

        # Pseudocount
        df_kmer_counts = df_kmer_counts + pseudocount

        # Normalize
        if frequency:
            return normalize(df_kmer_counts, method="tss", axis=1)
        else:
            return df_kmer_counts.astype(int)

# Prevalence
def prevalence(X:pd.DataFrame, interval_type="closed", name=None, scale=False) -> pd.Series:
    # Dimensions
    n,m = X.shape

    assert_acceptable_arguments(query=[interval_type],target=["closed", "open"], operation="le")

    # Attribtue depth
    data = (X > 0).sum(axis=0)

    # Comparison operation
    operation = {"closed":operator.ge, "open":operator.gt}[interval_type]

    # Number of attributes for tolerance
    d_tol_numattrs = OrderedDict()
    for tol in range(1,n+1):
        d_tol_numattrs[tol] = data.map(lambda x: operation(x,tol)).sum()

    # Prevalence
    x_prevalence = pd.Series(d_tol_numattrs, name=name)

    # Scale the prevalence
    if scale:
        x_prevalence.index = x_prevalence.index/n
    return x_prevalence



# ==============================================================================
# Modules
# ==============================================================================
# Built-ins
import os, sys, time, warnings
from collections import OrderedDict, defaultdict

# PyData
import pandas as pd
import numpy as np

# Compositional
# from skbio.stats.composition import clr, ilr
# from gneiss.composition import ilr_transform
import ete3
import compositional as coda

# SciPy
from scipy import stats
from sklearn.preprocessing import MinMaxScaler

# Soothsayer
from .conversion import ete_to_skbio
from ..r_wrappers.packages.edgeR import normalize_edgeR
from ..r_wrappers.packages.metagenomeSeq import normalize_css
# from soothsayer.r_wrappers.packages.philr import normalize_philr


from ..utils import is_dict, assert_acceptable_arguments, is_number

__all__ = {"normalize_minmax", "normalize_tss", "normalize_center", "normalize_zscore", "normalize_quantile", "normalize_boxcox",  "normalize", "normalize_expression"}

functions_from_compositional = {"transform_clr","transform_xlr", "transform_iqlr", "transform_ilr"}

for function_name in functions_from_compositional:
    globals()[function_name] = getattr(coda, function_name)
    __all__.add(function_name)

__all__ = sorted(__all__)
   

# ==============================================================================
# Normalization
# ==============================================================================
# Minmax
def normalize_minmax(X, feature_range=(0,1)):
    assert isinstance(X, (pd.Series, pd.DataFrame)), "X must be a pd.Series or pd.DataFrame"
    if isinstance(X, pd.Series):
        values = MinMaxScaler(feature_range=feature_range).fit_transform(X.values.reshape(-1,1)).ravel()
        return pd.Series(values, index=X.index, name=X.name)
    if isinstance(X, pd.DataFrame):
        return X.apply(lambda x: normalize_minmax(x, feature_range=feature_range), axis=1)

# Normalize by summing to one
def normalize_tss(X):
    """
    NumPy or Pandas
    """
    if len(X.shape) == 2: # if type(X) is pd.DataFrame:
        sum_ = X.sum(axis=1)
        return (X.T/sum_).T
    if len(X.shape) == 1: #elif type(X) is pd.Series:
        return X/X.sum()

# Center Normalization
def normalize_center(X):
    if isinstance(X, pd.DataFrame):
        A = X.values
        return pd.DataFrame(A - np.nanmean(X, axis=1).reshape((-1,1)), index=X.index, columns=X.columns)
    elif isinstance(X, pd.Series):
        return X - X.mean()

# Zscore Normalization
def normalize_zscore(X):
    if isinstance(X, pd.DataFrame):
        A_center = normalize_center(X).values
        A_std = np.nanstd(X.values, axis=1).reshape((-1,1))
        return pd.DataFrame(A_center/A_std, index=X.index, columns=X.columns)
    elif isinstance(X, pd.Series):
        return (X - X.mean())/X.std()

# Quantile Normalization
def normalize_quantile(X):
    assert isinstance(X, pd.DataFrame), "Quantile normalization is only compatible with `pd.DataFrame`"
    X_transpose = X.T
    X_mu = X_transpose.stack().groupby(X_transpose.rank(method='first').stack().astype(int)).mean()
    return X_transpose.rank(method='min').stack().astype(int).map(X_mu).unstack().T

# Boxcox Normalization
def normalize_boxcox(X):
    if isinstance(X, pd.DataFrame):
        return X.apply(lambda x:stats.boxcox(x)[0], axis=1)
    elif isinstance(X, pd.Series):
        return pd.Series(stats.boxcox(X)[0],index=X.index, name=X.name)

# Normalization
def normalize(X, method="tss", axis=1, tree=None, feature_range=(0,1)):
    """
    Assumes X_data.index = Samples, X_data.columns = Attributes
    axis = 0 = cols, axis = 1 = rows
    e.g. axis=1, method=ratio: Normalize for relative abundance so each row sums to 1.
    "tss" = total-sum-scaling
    "center" = mean centered
    "clr" = center log-ratio
    "quantile" = quantile normalization
    "zscore" = zscore normalization (mu: 0, var:1)
    "boxcox"
    "ilr" = isometric log ratio transform
    """
    if method in ["ratio", "relative-abundance"]:
        method = "tss"
    # Transpose for axis == 0
    if axis == 0:
        X = X.T
    # Common
    if method == "tss":
        df_normalized = normalize_tss(X)
    if method == "center":
        df_normalized = normalize_center(X)
    if method == "zscore":
        df_normalized = normalize_zscore(X)
    if method == "quantile":
        df_normalized = normalize_quantile(X)
    if method == "boxcox":
        df_normalized = normalize_boxcox(X)
    if method == "minmax":
        df_normalized = normalize_minmax(X, feature_range=feature_range)
    # Aitchison
    if method == "clr":
        df_normalized = transform_clr(X)
    if method == "ilr":
        df_normalized = transform_ilr(X, tree=tree)
    # if method == "philr":
    #     df_normalized = normalize_philr(X, tree=tree)

    # Transpose back
    if axis == 0:
        df_normalized = df_normalized.T
    return df_normalized

# Normalize gene expresssion data
def normalize_expression(X:pd.DataFrame, method:str="tpm", length:pd.Series=None, p=0.75, kws=dict()):
    """
    # FPKM
    Fragments Per Kilobase of transcript per Million mapped reads
        C = Number of reads mapped to a gene
        N = Total mapped reads in the experiment
        L = exon length in base-pairs for a gene
        Equation: RPKM = (10^9 * C)/(N * L)
            or
        numReads / ( geneLength/1000 * totalNumReads/1,000,000 ) # C/((L/1e3).T*(N/1e6)).T

    # TPM
    Transcripts Per Kilobase Million
        (1) Divide the read counts by the length of each gene in kilobases. This gives you reads per kilobase (RPK).
        (2) Count up all the RPK values in a sample and divide this number by 1,000,000. This is your “per million” scaling factor.
        (3) Divide the RPK values by the “per million” scaling factor. This gives you TPM.

    # TMM
    https://genomebiology.biomedcentral.com/articles/10.1186/gb-2010-11-3-r25

    # GeTMM
    https://bmcbioinformatics.biomedcentral.com/articles/10.1186/s12859-018-2246-7#MOESM4

    # RLE
    https://genomebiology.biomedcentral.com/articles/10.1186/gb-2010-11-10-r106

    # Upperquartile
    https://bmcbioinformatics.biomedcentral.com/articles/10.1186/1471-2105-11-94

    # CSS
    https://github.com/biocore/qiime/blob/6d8ffca3ffe2bb40057b8f9cbc08e2029b96d790/qiime/support_files/R/CSS.r

    # Notes:
    Proper pseudocount addition would be the following as shown by metagenomeSeq's MRcounts
    log(X_normalized + 1)
    """
    if method is None:
        return X

    method = method.lower()
    assert_acceptable_arguments(query=[method], target={"rpk", "fpkm", "rpkm", "tpm", "tmm", "getmm", "rle", "upperquartile", "css"})


    # Lengths
    if method in {"fpkm", "rpkm", "rpk",  "tpm", "getmm"}:
        assert length is not None, "If FPKM, RPKM, TPM, or GeTMM is chosed as the method then `length` cannot be None.  It must be either a pd.Series of sequences or sequence lengths"
        length = pd.Series(length)[X.columns]
        assert length.isnull().sum() == 0, "Not all of the genes in `X.columns` are in `length.index`.  Either use a different normalization or get the missing sequence lengths"
        # If sequences are given then convert to length (assumes CDS and no introns)
        if pd.api.types.is_string_dtype(length):
            length = length.map(len)

    # FPKM, RPKM, and TPM normalization
    if method in {"fpkm", "rpkm", "rpk", "tpm"}:

        # Set up variables
        C = X.values
        L = length.values
        N = X.sum(axis=1).values.reshape(-1,1)


        if method in {"fpkm","rpkm"}:
            # Compute operations
            numerator = 1e9 * C
            denominator = (N*L)
            return pd.DataFrame(numerator/denominator, index=X.index, columns=X.columns)

        if method in {"rpk", "tpm"}:
            rpk = C/L
            if method == "rpk":
                return pd.DataFrame(rpk, index=X.index, columns=X.columns)
            if method == "tpm":
                per_million_scaling_factor = (rpk.sum(axis=1)/1e6).reshape(-1,1)
                return pd.DataFrame( rpk/per_million_scaling_factor, index=X.index, columns=X.columns)


    # TMM/GeTMM Normalization
    if method in {"tmm", "getmm", "rle", "upperquartile"}:
        return normalize_edgeR(X, length=length, method=method, p=p, **kws)
    if method in  {"css"}:
        return normalize_css(X)

# # CLR Normalization
# def transform_clr(X, return_zeros_as_neginfinity=False):
#     """
#     Extension of CLR from skbio to handle zeros and NaN; though, this implementation will be slightly slower.

#     Please refer to the following for CLR documention:
#     http://scikit-bio.org/docs/latest/generated/skbio.stats.composition.clr.html#skbio.stats.composition.clr

#     """
#     assert isinstance(X, (pd.Series, pd.DataFrame)), "Type must be either pd.Series or pd.DataFrame"
#     # X
#     if isinstance(X, pd.DataFrame):

#         return transform_xlr(X, centroid="log_mean",return_zeros_as_neginfinity=return_zeros_as_neginfinity )

#     # x
#     if isinstance(X, pd.Series):
#         # Get values
#         X_values = X.values
#         # Check for zeros
#         X_contains_zeros = False
#         num_zeros = np.any(X == 0).flatten().sum()
#         if num_zeros:
#             warnings.warn("N={} detected in `X`.  Masking them as NaN to perform nan-robust functions".format(num_zeros))
#             X_zero_mask = X == 0
#             X_values[X_zero_mask] = np.nan
#             X_contains_zeros = True

#         # Transform
#         X_log = np.log(X)
#         centroid = np.nanmean(X_log, axis=-1)
#         X_transformed = X_log - centroid

#         # Output
#         if all([return_zeros_as_neginfinity, X_contains_zeros]):
#             X_transformed[X_zero_mask] = -np.inf
#         return pd.Series(X_transformed,index=X.index, name=X.name)

# # ===========================
# # Compositional data analysis
# # ===========================
# # Extension of CLR to use different centroid, references, and zeros without pseudocounts
# def transform_xlr(X:pd.DataFrame, reference_components=None, centroid="log_mean", return_zeros_as_neginfinity=False):
#     """
#     Extension of CLR to incorporate custom metrics such as median and harmonic mean.
#     If you want CLR, please use skbio's implementation as it is faster.
#     This implementation is more versatile with more checks but that makes it slower if it done iteratively.

#     This was designed to handle zero values.  It computes the centroid metrics for all non-zero values (masked as nan)
#     and returns them either as nan (can be used with some correlation functions) or as -inf (for mathematical consistency)

#     Documentation on CLR:
#     http://scikit-bio.org/docs/latest/generated/skbio.stats.composition.clr.html#skbio.stats.composition.clr

#     centroid: {log_median, log_mean, log_hmean, Type[Int]} If int, then that value is used as a percentile (e.g. median is 50)

#     Note: log_mean == arithmetic mean of logs == log of geometric mean
#     """

#     assert np.all(X >= 0), "`X` cannot contain negative values because of log-transformation step."
#     assert len(X.shape) == 2, "`X` must be 2-Dimensional"

#     # Check for labels
#     index = None
#     columns = None
#     X_values = X.astype(float)
#     X_is_labeled = False
#     if isinstance(X, pd.DataFrame):
#         index = X.index
#         columns = X.columns
#         X_values = X.values.astype(float)
#         if reference_components is None:
#             reference_components = columns
#         reference_components = list(map(lambda component: columns.get_loc(component), reference_components))
#         X_is_labeled = True
        
#     # Check centroid
#     if X_is_labeled:
#         if not isinstance(centroid, str):
#             if is_dict(centroid):
#                 centroid = pd.Series(centroid)
#             assert isinstance(centroid, pd.Series), "If `centroid` is dict-like/pd.Series then `X` must be a `pd.DataFrame`."
#             assert set(centroid.index) >= set(index), "Not all indicies from `centroid` are available in `X.index`."
#             centroid = centroid[index].values

#     # Check for zeros
#     X_contains_zeros = False
#     num_zeros = np.any(X == 0).flatten().sum()
#     if num_zeros:
#         warnings.warn("N={} detected in `X`.  Masking them as NaN to perform nan-robust functions".format(num_zeros))
#         X_zero_mask = X == 0
#         X_values[X_zero_mask] = np.nan
#         X_contains_zeros = True

#     # Log transformation
#     X_log = np.log(X_values)

#     # Centroid precomputed
#     if isinstance(centroid, str) or is_number(centroid):

#         # Centroid reference
#         median = {"log_median"}
#         geometric_mean = {"log_mean", "gmean_log" }
#         harmonic_mean = {"log_hmean"}

#         if isinstance(centroid, str):
#             centroid = centroid.lower()
#             assert_acceptable_arguments(centroid, median  | geometric_mean | harmonic_mean)

#             if centroid in median:
#                 centroid = np.nanmedian(X_log[:,reference_components], axis=-1)
#             elif centroid in geometric_mean:
#                 centroid = np.nanmean(X_log[:,reference_components], axis=-1)
#             elif centroid in harmonic_mean:
#                 if X_contains_zeros:
#                     centroid = np.asarray(list(map(lambda x:stats.hmean(x[np.isfinite(x)]), X_log[:,reference_components])))
#                 else:
#                     centroid = stats.hmean(X_log[:,reference_components], axis=-1)

#         # Percentiles
#         if is_number(centroid):
#             centroid = np.asarray(list(map(lambda x:np.percentile(x, centroid), X_log[:,reference_components])))
#     else:
#         assert any([is_dict_like(centroid), np.ndarray]), "If centroid is not a string or a numeric value then it must be either a dict-like/pd.Series (labeled) or numpy array (unlabeled)"

#     # Check dimensions
#     assert len(centroid) == X_log.shape[0], "Dimensionality is not compatible: centroid.size != X.shape[0]."
#     centroid = np.asarray(centroid).reshape(-1,1)
#     # Transform
#     X_transformed = X_log - centroid

#     # Output
#     if all([return_zeros_as_neginfinity, X_contains_zeros]):
#         X_transformed[X_zero_mask] = -np.inf

#     if X_is_labeled:
#         return pd.DataFrame(X_transformed, index=index, columns=columns)
#     else:
#         return X_transformed

# # Interquartile range log-ratio transform
# def transform_iqlr(X:pd.DataFrame, percentile_range=(25,75), centroid="log_mean", interval_type="open", return_zeros_as_neginfinity=False):
#     """
#     Interquartile range log-ratio transform
#     interval: open = (a,b) and closed = [a,b].  open is used by `propr` R package:
#     https://github.com/tpq/propr/blob/2bd7c44bf59eaac6b4d329d38afd40ac83e2089a/R/2-proprCall.R#L31
#     """
#     assert_acceptable_arguments(query=[interval_type],target=["closed", "open"], operation="le")
#     percentile_range = tuple(sorted(percentile_range))
#     assert len(percentile_range) == 2, "percentile_range must have 2 elements"
#     assert isinstance(X, pd.DataFrame), "`transmute_iqlr` can currently only support pd.DataFrames and not np.arrays"
#     # Compute the variance of the CLR transform
#     clr_var = transform_xlr(X, centroid=centroid).var(axis=0)
#     # Remove non-finite values
#     clr_var = clr_var[np.isfinite(clr_var)]
#     # Calculate upper and lower bounds from percentiles
#     lower_bound, upper_bound = np.percentile(clr_var, percentile_range)
#     # Get the reference components
#     if interval_type == "open":
#         reference_components = clr_var[(lower_bound < clr_var) & (clr_var < upper_bound)].index
#     if interval_type == "closed":
#         reference_components = clr_var[(lower_bound <= clr_var) & (clr_var <= upper_bound)].index
#     return transform_xlr(X, reference_components=reference_components, centroid=centroid, return_zeros_as_neginfinity=return_zeros_as_neginfinity)

# # ILR Transformation
# def transform_ilr(X:pd.DataFrame, tree=None, node_prefix="y", bifurcation_kws=dict(), verbose=True):
#     """
#     if `tree` is None then orthonormal basis for Aitchison simplex defaults to J.J.Egozcue orthonormal basis.
#     """
#     assert isinstance(X, pd.DataFrame), "type(X) must be pd.DataFrame"
#     assert not np.any(X == 0), "`X` cannot contain any zero values because of the log-transform.  Give it a pseudocount. (X+1)"

#     # Supply tree
#     if tree is not None:
#         tree_set =  set(tree.get_leaf_names())
#         leaf_set_from_X = set(X.columns)
#         assert leaf_set_from_X <= tree_set, "`X` columns should be a subset of `tree` leaves"
#         tree = tree.copy(method="deepcopy")
#         if leaf_set_from_X < tree_set:
#             n_before_pruning = len(tree_set)
#             tree.prune(leaf_set_from_X)
#             if verbose:
#                 print(f"Pruned {n_before_pruning - len(tree.get_leaves())} attributes to match X.columns", file=sys.stderr)

#         if isinstance(tree, (ete3.Tree, ete3.PhyloTree, ete3.ClusterTree)):
#             # Check bifurcation
#             n_internal_nodes = len([*filter(lambda node:node.is_leaf() == False, tree.traverse())])
#             n_leaves = len([*filter(lambda node:node.is_leaf(), tree.traverse())])
#             if n_internal_nodes < (n_leaves - 1):
#                 tree.resolve_polytomy(**bifurcation_kws)
#                 if verbose:
#                     print("Resolving tree polytomy and forcing bifurcation", file=sys.stderr)
#             tree = ete_to_skbio(tree=tree, node_prefix=node_prefix)
#         return ilr_transform(table=X, tree=tree)
#     else:
#         return pd.DataFrame(ilr(X), index=X.index)



import sys 
__all__ = list()
# if "rpy2" in sys.modules:
    # from rpy2.robjects.packages import importr
    # from rpy2 import __version__ as rpy2_version
    # rpy2_version_major = int(rpy2_version.split(".")[0])
    # assert rpy2_version_major > 1, "Please update your rpy2 version"
    # if rpy2_version_major == 2:
    #     from rpy2.rinterface import RRuntimeError
    #     importing_error = RRuntimeError
    # if rpy2_version_major == 3:
    #     from rpy2.robjects.packages import PackageNotInstalledError
    #     importing_error = PackageNotInstalledError
__all__ += ["R_package_retrieve", "R_packages", "pandas_to_rpy2", "rpy2_to_pandas"]
from .r_wrappers import *
from .packages import dynamicTreeCut
from .packages import WGCNA
from .packages import philr
from .packages import edgeR
from .packages import metagenomeSeq
from .packages import propr

__all__ += ["dynamicTreeCut", "WGCNA", "philr", "edgeR", "metagenomeSeq" ,"propr"]

__all__ = sorted(__all__)



# ==============================================================================
# Imports
# ==============================================================================
# Built-ins
import os, sys
# ==============================================================================
# R Compatibility
# ==============================================================================
from ..utils import check_packages
if "rpy2" in sys.modules:
    from rpy2 import robjects as ro
    from rpy2 import rinterface as ri
    from rpy2.robjects import pandas2ri
    from rpy2 import __version__ as rpy2_version
    rpy2_version_major = int(rpy2_version.split(".")[0])
    rpy2_version_minor = int(rpy2_version.split(".")[1])
    # rpy2_version_micro = int(rpy2.__version__[2])
    assert rpy2_version_major > 1, "Please update your rpy2 version"


# Converters
@check_packages(["rpy2"], language="python", import_into_backend=False)
def pandas_to_rpy2(df, rpy2_version_major=None):
    if rpy2_version_major is None:
        from rpy2 import __version__ as rpy2_version
        rpy2_version_major = int(rpy2_version.split(".")[0])
    if rpy2_version_major == 2: # v2.x
        return pandas2ri.py2ri(df)
    if rpy2_version_major == 3: # v3.x
        with localconverter(ro.default_converter + pandas2ri.converter):
            return ro.conversion.py2rpy(df)

@check_packages(["rpy2"], language="python", import_into_backend=False)
def rpy2_to_pandas(r_df, rpy2_version_major=None):
    if rpy2_version_major is None:
        from rpy2 import __version__ as rpy2_version
        rpy2_version_major = int(rpy2_version.split(".")[0])
    if rpy2_version_major == 2: # v2.x
        return pandas2ri.ri2py(r_df)
    if rpy2_version_major == 3: # v3.x
        with localconverter(ro.default_converter + pandas2ri.converter):
            return ro.conversion.rpy2py(r_df)

# Load R Packages
R_packages = dict() # Remove this in future versions
@check_packages(["rpy2"], language="python", import_into_backend=False)
def R_package_retrieve(package, not_exists_ok=True):
    """
    ["dynamicTreeCut", "WGCNA", "fastcluster", "phyloseq", "philr", "ape", "metagenomeSeq", "edgeR"]
    """
    from rpy2.robjects.packages import importr
    if rpy2_version_major == 2:
        from rpy2.rinterface import RRuntimeError
        importing_error = RRuntimeError
        pandas2ri.activate()
        ri.set_writeconsole_regular(None) # How do I do this v3?
    if rpy2_version_major == 3:
        # from rpy2.rinterface_lib.embedded import RRuntimeError
        from rpy2.robjects.conversion import localconverter
        from rpy2.robjects.packages import PackageNotInstalledError
        importing_error = PackageNotInstalledError
    try:
        return R_packages[package]
    except KeyError:
        try:
            R_packages[package] = importr(package)
            return R_packages[package]
        except importing_error:
            msg = f"{package} is not available"
            if not_exists_ok:
                print(msg, file=sys.stderr)
            else:
                raise ImportError(msg)










#



# Built-ins
import os, sys, itertools, time, copy, logging
from collections import OrderedDict, defaultdict, Mapping

# PyData
import pandas as pd
import numpy as np
import xarray as xr

## Plotting
import matplotlib.pyplot as plt
import seaborn as sns

# Machine learning
from scipy import stats
from sklearn import model_selection, metrics
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier
# from sklearn.multiclass import OneVsRestClassifier
from sklearn.base import clone

# Parallel
from joblib import Parallel, delayed

# Utility
from tqdm import tqdm

# Soothsayer
from soothsayer.utils import *
from soothsayer.io import read_object, write_object

# Clairvoyant
class Clairvoyant(object):
    """
    Clairvoyant is an algorithm that weights attributes that can optimize the predictive capacity of a dataset

    devel
    ============
    # 2018-April-18
    (1) Added decision tree option
    (2) Added repr for relevant metadata and parameters
    (3) Added a metadata container
    (4) Added an indexable pd.DataFrame with all of the model indicies and the accuracies from the intial fitting
    (5) Added default base_model for cross-validation using hyperparameters with best accuracy from initial fitting
    (6) Adjusted (5) to sort by accuracy and if there are multiple equally good hyperparameters, it favors settings closer to the default settings
    (7) When min_threshold is None it calculates the weights from `threshold_range` (100 linear spaced segments)
    # 2018-April-24
    (8) Added standard error of the mean in the plots
    (9) Added sem to scores output
    (10) Added attribute set to scores output
    (11) Fixed a HUGE error where the X.index and y.index were getting offset because I was looking for overlaps
    (12) Added functionality for cross-validation to include a target_score.  Mostly useful for commandline version
    # 2018-June-01
    (13) Added value of min_threshold in the weights
    (14) Added a default scaling for weights in mode=0 and mode=2 so it does not need to be done in mode=1
    (15) Removing n_space and making it default number of hyperparameter configurations for both logistic and tree (n=40)
    # 2018-August-22
    (16) Added early_stopping
    (17) Added weights for specific classes for LogisticRegression (tree-based models will be next)
    # 2018-October-08
    (18) Fixed error introduced from process tuples as indicies in pandas 0.23.4: https://github.com/pandas-dev/pandas/issues/22832

    Previous Versions
    =================
    seeker
    seeker_linear

    Future
    ======
    Refine option for softmax and collapsing groups
    Set up models to work on any combination of parameters
    Add other models (SVC?)
    Randomly sample attributes

    Benchmarks
    ==========
    sklearn.__version__ = 0.19.1
    (No early_stopping)
    Dataset: (150, 54)
    LogisticRegression:
    Fitting = 2.66 ms ± 111 µs per loop (mean ± std. dev. of 7 runs, 1 loop each)
    Predicting = 82.2 µs ± 214 ns per loop (mean ± std. dev. of 7 runs, 10000 loops each)

    DecisionTreeClassifier:
    Fitting = 1.7 ms ± 14.5 µs per loop (mean ± std. dev. of 7 runs, 1000 loops each)
    Predicting = 81.5 µs ± 1.26 µs per loop (mean ± std. dev. of 7 runs, 10000 loops each)

    Notes
    ====
    Note, only set up for specific parameters (i.e. only use C,penalty for lr and criterion,max_features,min_samples_leaf for tree)
    """
    def __init__(
        self,
        # Base
        name=None,
        # Iteration
        n_iter=10,
        # Modeling
        random_state=0,
        model_type="logistic",
        n_jobs=-1,
        random_mode=0,
        # Logistic Regression
        lr_space={
            "C":np.linspace(1e-10,1.0,20),
            "penalty":["l1", "l2"],
        },
        lr_kws=dict(max_iter=10000, solver="liblinear"),
        # Decision Tree
        tree_space={
            "criterion":["gini","entropy"],
            "max_features":["log2", "sqrt", None, 0.382],
            "min_samples_leaf":[1,2,3,5,8],
        },
        tree_kws={
        },
        # Labeling
        map_encoding=None,
        attr_type="attr",
        class_type="class",
        verbose=True,
        metadata=dict(),
    ):
        """
        Formula for `shape_output`:
        n_models = 2*n_iter*40parameterconfigs = 80*n_iter

        Default values result in 40 parameter configurations, fitted on set A and B, resulting in 80 models per iteration.


        """
        self.name = name
        self.n_iter = n_iter
        if isinstance(random_state, int):
            random_state = np.random.RandomState(int(random_state))
        self.random_state= random_state
        self.random_state_value = random_state.get_state()[1][0]
        self.random_mode = random_mode
        # Logistic Regression
        self.lr_space = copy.deepcopy(lr_space)
        self.lr_kws = lr_kws
        # Decision Tree
        self.tree_space = copy.deepcopy(tree_space) # May not need the deepcopy here
        self.tree_kws = tree_kws
        self.model_type = model_type.lower()
        self.models = self._placeholder_models()
        self.n_jobs = n_jobs
        self.map_encoding = map_encoding
        self.attr_type = attr_type
        self.class_type = class_type
        self.verbose = verbose
        # Metadata
        self.metadata = {
            "name":self.name,
            "n_models":self.n_models,
            "model_type":self.model_type
        }
        self.metadata.update(metadata)

    def __repr__(self):
        class_name = self.__class__.__name__
        return f"{class_name}({str(self.metadata)[1:-1]})"
    def copy(self):
        return copy.deepcopy(self)

    # =================================================================================
    # Fitting
    # =================================================================================
    def _placeholder_models(self):
        """
        Internal: Label models
        Need to extend this for using different hyperparameters
        """
        self.param_index = OrderedDict()
        def _get_random_state(mode, i):
            d_random = {
                0: self.random_state,
                1: np.random.RandomState(self.random_state_value+i),
                2: np.random.RandomState(np.random.randint(low=0, high=int(1e5)))
            }
            return d_random[mode]

        # Indexing for hyperparameters and models
        models = list()
        a = np.ones(self.n_iter*2)
        b = np.arange(a.size)
        if self.model_type == "logistic":
            assert all([*map(lambda x: x in self.lr_space, ["penalty", "C"])]), "`lr_kws` must contain `penalty` and `C`"
            param_set = [*map(lambda x:dict(zip(self.lr_space.keys(), x)), itertools.product(*self.lr_space.values()))]
            c = np.asarray(len(param_set)*a*b, dtype=int)
            for i, hyperparameters in enumerate(param_set):
                # Random State
                rs = _get_random_state(mode=self.random_mode, i=i)
                # Construct model
                model = LogisticRegression(random_state=rs, multi_class="ovr", **hyperparameters, **self.lr_kws)
                models.append(model)
                # Store the resulting indicies that will correspond with these hyperparameters
                self.param_index[tuple(hyperparameters.items())] = c + i

        if self.model_type == "tree":
            assert all([*map(lambda x: x in self.tree_space, ["criterion", "max_features", "min_samples_leaf"])]), "`tree_kws` must contain `criterion`, `max_features`, and `min_samples_leaf`"
            param_set = [*map(lambda x:dict(zip(self.tree_space.keys(), x)), itertools.product(*self.tree_space.values()))]
            c = np.asarray(len(param_set)*a*b, dtype=int)
            for i, hyperparameters in enumerate(param_set):
                # Random State
                rs = _get_random_state(mode=self.random_mode, i=i)
                # Construct model
                model = DecisionTreeClassifier(random_state=rs, **hyperparameters, **self.tree_kws)
                models.append(model)
                # Store the resulting indicies that will correspond with these hyperparameters
                self.param_index[tuple(hyperparameters.items())] = c + i
        self.n_models = len(models)*self.n_iter*2
        return models

    def _fit_partitions(self, X_A, X_B, y_A, y_B, model, weight_type):
        """
        Internal: Get coefs
        accuracy_A_B means accuracy of B trained on A
        """
#         print(X_A.shape, X_B.shape, y_A.shape, y_B.shape, model, weight_type)
        # Subset A
        model.fit(X_A, y_A)
        weight_A = getattr(model, weight_type)
        accuracy_A_B = np.mean(model.predict(X_B) == y_B)
        # Subset B
        model.fit(X_B, y_B)
        weight_B = getattr(model, weight_type)
        accuracy_B_A = np.mean(model.predict(X_A) == y_A)
        return [weight_A, weight_B], [accuracy_A_B, accuracy_B_A]

    def _merge_partitions(self, X, y, stratify, subset, desc):
        """
        Internal: Distribute tasks
        # (n_space_per_iteration, 2_splits, n_classifications, n_attributes)
        """

        if self.verbose:
            iterable = tqdm(range(self.random_state_value, self.random_state_value + self.n_iter), desc=desc)
        else:
            iterable = range(self.random_state_value, self.random_state_value + self.n_iter)

        # Iterate through n_iter
        weights = list()
        accuracies = list()

        for i in iterable:
            rs = np.random.RandomState(i)
            X_A, X_B, y_A, y_B = model_selection.train_test_split(X,y,test_size=subset,stratify=stratify, random_state=rs)
            params = {"X_A":X_A, "X_B":X_B, "y_A":y_A, "y_B":y_B}
            parallel_results = Parallel(n_jobs=self.n_jobs)(delayed(self._fit_partitions)(**params,
                                                                                          model=model,
                                                                                          weight_type={"logistic":"coef_","tree":"feature_importances_"}[self.model_type]) for model in self.models)

            # Extract weights and accuracy
            # Convert to array and collapse (X_A and X_B)
            # Store the weights and accuracies

            # Weights
            if self.model_type == "logistic":
                # Might be able to condense the w for both
                w = np.concatenate([*map(lambda x: x[:-1], parallel_results)], axis=0) #w.shape ==> (n_space_per_iteration, 2_splits, n_classifications, n_attributes)
                w = np.concatenate([w[:,0,:,:], w[:,1,:,:]], axis=0) # w shape:  (2*n_space_per_iteration, n_classifications, n_attributes)
            if self.model_type == "tree":
                w = np.asarray([*map(lambda x: x[0], parallel_results)])
                w = np.concatenate([w[:,0], w[:,1]], axis=0)
            weights.append(w)
            # Accuracies
            acu = np.asarray([*map(lambda x: x[-1], parallel_results)]) #acu shape ==> (n_space_per_iteration, 2_splits)
            acu = np.concatenate([acu[:,0], acu[:,1]], axis=0) #acu shape ==> (2*n_space_per_iteration, )
            accuracies.append(acu)

        # Merge all of the weights and accuracies
        weights = np.concatenate(weights, axis=0)
        accuracies = np.concatenate(accuracies, axis=0)

        return (weights, accuracies)

    def fit(self, X, y, stratify=True, subset=0.5, desc="Permuting samples and fitting models"):
        """
        Fit data in parallel. Rewrite this to enable loading data.
        """
        self.X = X.copy()
        self.y = y.copy()
        self.obsv_ids = self.X.index
        self.attr_ids = self.X.columns
        self.shape_input_ = self.X.shape
        self.class_labels = sorted(y.unique())
        self.class_labels_encoding = self.class_labels
        if self.map_encoding is not None:
            self.class_labels = [*map(lambda x:self.map_encoding[x], self.class_labels)]
        if stratify:
            stratify = self.y
        (self.kernel_, self.accuracies_) = self._merge_partitions(X=self.X, y=self.y, stratify=stratify, subset=subset, desc=desc)

        self.shape_ = self.kernel_.shape
        max_accuracy = self.extract_hyperparameters()["accuracy"].max()
        idx_maxacu = self.extract_hyperparameters()["accuracy"][lambda x: x == max_accuracy].index
        if len(idx_maxacu) > 1:
            if self.model_type == "logistic":
                # Prefer l2 with higher C
                self.best_hyperparameters_ = pd.DataFrame([*map(dict,idx_maxacu)]).sort_values(["penalty", "C"], ascending=[False,False]).iloc[0,:].to_dict()
            elif self.model_type == "tree":
                # Prefer lower min_samples_leaf, gini, and sqrt
                self.best_hyperparameters_ = pd.DataFrame([*map(dict,idx_maxacu)]).sort_values(["min_samples_leaf", "criterion", "max_features"], ascending=[True,False, False]).iloc[0,:].to_dict()
        else:
            self.best_hyperparameters_ = dict(idx_maxacu[0])
        self.metadata["kernel_shape"] = self.shape_
        self.metadata["best_hyperparameters"] = self.best_hyperparameters_
        return self


    # =================================================================================
    # Extracting
    # =================================================================================
    def extract_kernel(self, into="numpy", name=None, min_threshold=None):
        """
        Extract coefficients from training
        if penalty is specific, an array for l1 or l2 loss specifically is output.  Does not support xarray
        """
        min_threshold = 0 if min_threshold in [False, None] else min_threshold

        W = self.kernel_[self.accuracies_ > min_threshold]

        if into == "numpy":
            return W

        if self.model_type == "logistic":
            if into == "xarray":
                if W.shape[1] > 1:
                    return xr.DataArray(W, dims=["index_model", self.class_type, self.attr_type], coords=[range(W.shape[0]), self.class_labels, self.attr_ids], name=name)
                else:
                    return xr.DataArray(np.squeeze(W), dims=["index_model",  self.attr_type], coords=[range(W.shape[0]), self.attr_ids], name=name)
            if into == "pandas":
                panel = pd.Panel(W,
                             major_axis=pd.Index(range(W.shape[1]), level=0, name=self.class_type),
                             minor_axis=pd.Index(self.attr_ids, level=1, name=self.attr_type))
                df = panel.to_frame().T
                df.index.name = "index"
                return df
        elif self.model_type == "tree":
            if into == "xarray":
                return xr.DataArray(W, dims=["index_model", self.attr_type], coords=[range(W.shape[0]),  self.attr_ids], name=name)
            if into == "pandas":
                return pd.DataFrame(W, index=pd.Index(range(W.shape[0]), name="index_model"), columns=pd.Index(self.attr_ids, name=self.attr_type))

    def extract_accuracies(self, into="numpy", name=None):
        """
        Extract coefficients from training
        if penalty is specific, an array for l1 or l2 loss specifically is output.  Does not support xarray
        """
        if into == "numpy":
            return self.accuracies_
        if into == "pandas":
            return pd.Series(self.accuracies_, name=name)

    def extract_weights(self, min_threshold=None, ascending=False, func_reduce=np.sum, func_weights=None, name=None, mode=0, threshold_range=(0,0.9)):
        """
        Extract weights
        """
        if all([(mode == 0), (min_threshold is None)]):
            mode = 2
        if mode == 0:
            # Extract weights from the kernel
            W = np.abs(self.extract_kernel(into="numpy", min_threshold=min_threshold))

            if self.model_type == "logistic":
                weights_ = pd.Series(func_reduce(func_reduce(W, axis=0), axis=0), index=self.attr_ids, name=name)
            if self.model_type == "tree":
                weights_ = pd.Series(func_reduce(W, axis=0), index=self.attr_ids, name=name)
            if func_weights is not None:
                weights_ = func_weights(weights_)

            weights_ = weights_/weights_.sum()

            if ascending is None:
                return weights_
            else:
                return weights_.sort_values(ascending=ascending)
        if mode == 1:
            # Extract a range of weights from the kernel
            data = dict()
            threshold_values = np.linspace(*threshold_range,100)
            if min_threshold is not None:
                threshold_values = sorted(set(threshold_values) | set([min_threshold]))
            for i in threshold_values:
                w = self.extract_weights(min_threshold=i, ascending=None, mode=0)
                data[i] = w
            df_w = pd.DataFrame(data).T
            df_w.index.name = "threshold"
            df_w.columns.name = "weights"
            return df_w.sort_index()
        if mode == 2:
            # Reduce all of the weights into a single vector
            weights_ = func_reduce(self.extract_weights(ascending=ascending, mode=1), axis=0)
            weights_ = weights_/weights_.sum()
            if ascending is None:
                return weights_
            else:
                return weights_.sort_values(ascending=ascending)
        if mode == 3:
            # Mode 3: Returns specific weights for each class
            # Note, this only works for logistic regression.  Future versions will use OneVsRestClassifier with DecisionTree
            # model = OneVsRestClassifier(DecisionTreeClassifier(random_state=0), n_jobs=-1)
            # model.fit(X,y)
            # np.asarray([*map(lambda x: x.feature_importances_, model.estimators_)])

            A = np.abs(self.extract_kernel(into="numpy", min_threshold=min_threshold))
            A = func_reduce(A, axis=0)
            W = A/A.sum(axis=1).reshape(-1,1)
            return pd.DataFrame(W,
                                index=pd.Index(self.class_labels, name=self.class_type),
                                columns=pd.Index(self.attr_ids, name=self.attr_type)).T



    def extract_hyperparameters(self, ascending=None):
        index = pd.Index([*map(frozenset, self.param_index.keys())], dtype="category")
        data = [*map(np.asarray, self.param_index.values())]
        df = pd.Series(data, index=index).to_frame("index_model")
        try:
            Se_hyperparam_acu = df["index_model"].map(lambda x: np.mean(self.accuracies_[x]))
            df.insert(loc=0, column="accuracy", value=Se_hyperparam_acu)
            if ascending in [True, False]:
                df = df.sort_values("accuracy", ascending=ascending)

        except AttributeError:
            print("Must run `.fit` before accuracies can be added", file=sys.stderr)
        df.index.name = "hyperparameters"
        return df
    ## set_weights is DEPRACATED because information is collected during the fitting process
    # def set_weights(self, external=None, *args):
    #
    #     if external is not None:
    #         self.weights_ = external
    #     else:
    #         self.weights_ = self.extract_weights(*args)
    #     if self.verbose:
    #         print("Updating weights...", file=sys.stderr)

    # # Load fitted kernel
    # def load_kernel(self, path):
    #     print("Eventually...")

    # =================================================================================
    # Cross Validation
    # =================================================================================
    def _softmax(self, x):
        """Compute softmax values for each sets of scores in x."""
        return np.exp(x) / np.sum(np.exp(x), axis=0)

    def _cross_val_score_parallel(self, model, X, y, Ar_X, Ar_y, y_groupby, idx_tr, idx_te, func_groupby, class_labels_encoding):
        model.fit(Ar_X[idx_tr,:],Ar_y[idx_tr])
        df_proba = pd.DataFrame(model.predict_proba(Ar_X[idx_te,:]), index=X.index[idx_te], columns=class_labels_encoding)
        df_groupby = df_proba.groupby(func_groupby).sum().apply(softmax, axis=1)
        y_hat = df_groupby.apply(np.argmax, axis=1)
        y_true = y_groupby[y_hat.index]
        return metrics.accuracy_score(y_true, y_hat)


### REPLACE WITH GROUPBY?
#     def _parse_compound(self,x):
#         return "-".join(x.split("|__|")[0].split("-")[:-1])
#     def _parse_replicate(self,x):
#         return x.split("|__|")[0]

    def cross_validate(self, model="auto", cv=5, cv_labels=None, n_jobs=-1, method="bruteforce", adaptive_range = [0, 0.5, 1.0], adaptive_steps = [1, 10], early_stopping=100, verbose=True, func_groupby=None, min_threshold=None, path_save=None, target_score=0, desc="Cross validating model subsets", log_file=sys.stderr, log_prefix=""):
        """
        2018-02-15
            * Added `bruteforce`
        Notes:
            * Automatically called extract weights?
        """

        self.performance_ = None
        # Hardcoded placeholder
        scoring="accuracy"
        # Model
        if model == "auto":
            if self.model_type == "logistic":
                model = LogisticRegression(random_state=self.random_state, multi_class="ovr",  **self.best_hyperparameters_, **self.lr_kws)
            elif self.model_type == "tree":
                model = DecisionTreeClassifier(random_state=self.random_state, **self.best_hyperparameters_, **self.tree_kws)

        # If labels are given instead of numeric indices
        condition_A = isinstance(cv, int) == False
        condition_B = type(cv) != list
        if all([condition_A, condition_B]):
            data_sample = cv[0][0]
            if set(data_sample) <= set(self.X.index):
                cv_idx = list()
                for idx_tr, idx_te in cv:
                    cv_idx.append((idx_tr.map(lambda x:self.X.index.get_loc(x)), idx_te.map(lambda x:self.X.index.get_loc(x))))
                cv = cv_idx

        # Attribute iterables
        if method == "bruteforce":
            if self.verbose:
                iterable = tqdm(range(1, self.X.shape[1]), desc=desc)
            else:
                iterable = range(1, self.X.shape[1])
        if method == "adaptive":
            positions = (np.asarray(adaptive_range)*self.X.shape[1]).astype(int)
            iterable = list()
            for i in range(len(adaptive_range)-1):
                pos = np.arange(positions[i], positions[i+1], adaptive_steps[i])
                iterable += pos.tolist()
            iterable = sorted(set(iterable) - set([0]))
            if self.verbose:
                iterable = tqdm(np.asarray(iterable), desc=desc)
            else:
                iterable = np.asarray(iterable) # Is np.asarray necessary?
        # Weights
        weights_ = self.extract_weights(min_threshold=min_threshold, ascending=False )

        # Cross Validate
        cv_summary = defaultdict(dict)
        cv_results = dict()
        label_attrs_included = f"num_{str(self.attr_type)}_included"
        label_attrs_excluded = f"num_{str(self.attr_type)}_excluded"

        # Cross validation
        scores = list()
        Ar_y = self.y.values

        # Baseline
        baseline_score  = np.mean(model_selection.cross_val_score(model, X=self.X.values, y=Ar_y, cv=cv, n_jobs=n_jobs, scoring=scoring))
        early_stopping_counter = 0
        for i,pos in enumerate(iterable):
            idx_query_attrs = weights_.index[:pos]
            if idx_query_attrs.size > 0:
                Ar_X = self.X.loc[:,idx_query_attrs].values
                # Original Implementation
                if func_groupby is None:
                    cross_validation_results = model_selection.cross_val_score(model, X=Ar_X, y=Ar_y, cv=cv, n_jobs=n_jobs, scoring=scoring)
                    accuracy = np.mean(cross_validation_results)
                    sem = stats.sem(cross_validation_results)
                 # Extended to PO1 Project [DEPRACATED]
                else:
                    raise Exception("Scoring must be `accuracy` and `func_groupby` is not None")
                    # if func_groupby in ["compound", "replicate"]:
                        # func_groupby = {"compound": self._parse_compound, "replicate":self._parse_replicate}[func_groupby]
                    # # Future: Cleaner way to remove duplicate indices
                    # y_groupby = pd.Series(pd.Series(Ar_y, index=self.y.index.map(func_groupby)).to_dict(), name="y_groupby")
                    # args = dict(X=self.X, y=self.y, Ar_X=Ar_X, Ar_y=Ar_y, y_groupby=y_groupby, func_groupby=func_groupby, class_labels_encoding=self.class_labels_encoding)
                    # parallel_results = np.asarray(Parallel(n_jobs=n_jobs)(delayed(self._cross_val_score_parallel)(model=clone(model),idx_tr=idx_tr, idx_te=idx_te, **args) for idx_tr, idx_te in cv))
                    # accuracy = np.mean(parallel_results)

                # Update scores
                if i > 0:
                    condition_A = accuracy > max(scores)
                    condition_B = accuracy > target_score
                    if all([condition_A, condition_B]):

                        if log_file is not None:
                            entry = "\t".join([str(log_prefix), f"Baseline={to_precision(baseline_score)}", f"iteration={i}", f"index={pos}", f"{label_attrs_included}={idx_query_attrs.size}", f"Accuracy={to_precision(accuracy)}", f"∆={to_precision(accuracy - baseline_score)}", f"SEM={to_precision(sem)}"])
                            if type(log_file) == logging.Logger:
                                log_file.info(entry)
                            if self.verbose:
                                if hasattr(log_file, 'write'):
                                    print(entry, file=log_file)
                        if path_save is not None:
                            save_data = dict(model=model, X=Ar_X, y=Ar_y, cv=cv, idx_attrs=list(idx_query_attrs), idx_samples=list(self.X.index), n_jobs=n_jobs, scoring=scoring, random_state=self.random_state)
                            write_object(save_data, path_save, compression="infer")
                        # Reset the Counter
                        early_stopping_counter = 0
                    else:
                        early_stopping_counter += 1
                # Organize results for dataframe
                cv_summary[pos][scoring] = accuracy
                cv_summary[pos]["sem"] = sem
                cv_summary[pos][label_attrs_included] = idx_query_attrs.size
                cv_summary[pos][f"{self.attr_type}_set"] = list(idx_query_attrs)

                # Individual cross-validations
                cv_results[pos] = pd.Series(cross_validation_results, index=[*map(lambda x:f"cv={x}", range(1, len(cross_validation_results)+1))])

                # Append scores
                scores.append(accuracy)

                if early_stopping_counter >= early_stopping:
                    if log_file is not None:
                        entry = f"Early stopping at iteration={i} with maximum score= {max(scores)}"
                        if type(log_file) == logging.Logger:
                            log_file.info(entry)
                        if hasattr(log_file, 'write'):
                            if self.verbose:
                                print(entry, file=log_file)
                    break
        # DataFrame for number of attributes included
        self.performance_ = pd.DataFrame(cv_summary).T
        self.performance_[label_attrs_included] = self.performance_[label_attrs_included].astype(int)
        self.performance_[label_attrs_excluded] = (self.X.shape[1] - self.performance_[label_attrs_included]).astype(int)
        self.performance_ = self.performance_.loc[:,[scoring, "sem", label_attrs_included, label_attrs_excluded, f"{self.attr_type}_set"]]

        # DataFrame for individual cross-validation results
        df_cvresults = pd.DataFrame(cv_results).T

        if cv_labels is not None:
            df_cvresults.columns = cv_labels
        self.performance_ = pd.concat([self.performance_, df_cvresults], axis=1)

        # Best attributes
        self.best_attrs_ = self.performance_.sort_values(["accuracy", "sem"], ascending=[False, True]).iloc[0,:][f"{self.attr_type}_set"]

        return self.performance_

    # =================================================================================
    # Plotting
    # =================================================================================
    def plot_scores(self, min_threshold=None, style="seaborn-white", figsize=(21,5), ax=None, size_scatter=50, ylabel=None,  title=None, title_kws=dict(), legend_kws=dict(), ylim=None, baseline_score=None):

        with plt.style.context(style):
            df_performance_wrt_included = self.performance_.set_index( f"num_{str(self.attr_type)}_included", drop=True)
            weights_ = self.extract_weights(min_threshold=min_threshold)
            scores = df_performance_wrt_included["accuracy"]
            df_cvresults = df_performance_wrt_included.iloc[:,4:]
            scores = df_cvresults.mean(axis=1)
            sems = df_cvresults.sem(axis=1)

            # Keywords
            _title_kws = dict(fontsize=15, fontweight="bold")
            _title_kws.update(title_kws)

            legend_fontsize = 8 if df_cvresults.shape[1] > 20 else 10
            _legend_kws = {'bbox_to_anchor': (1, 0.5), 'edgecolor': 'black', 'facecolor': 'white', 'fontsize': legend_fontsize, 'frameon': True, 'loc': 'center left'}
            _legend_kws.update(legend_kws)

            if ax is None:
                fig, ax = plt.subplots(figsize=figsize)
            else:
                fig = plt.gcf()


            # Get maximum indicies
            idx_maximums= list()
            values_maximums = list()

            current_maximum = 0
            for i,v in scores.iteritems():
                if v > current_maximum:
                    idx_maximums.append(i)
                    values_maximums.append(v)
                    current_maximum = v


            # Plot each cv accuracy
            df_cvresults.plot(color=sns.color_palette(palette="hls", n_colors=df_cvresults.shape[1]), ax=ax, alpha=0.382)
            # Plot mean accuracy
            ax.plot(scores.index, scores.values, color="black", label="", linewidth=1.618, alpha=0.618)
            # Plot progressive steps
            ax.scatter(idx_maximums, values_maximums, s=size_scatter, color="darkslategray", marker="o", label="", edgecolor="red", linewidth=1.618, alpha=1)
            # Accuracy is the only scoring metric available for the time being
            score_metric = "Accuracy" #str(self.performance_.columns[0]).capitalize()

            # Index of best score
            df_stats = df_performance_wrt_included.loc[:,["accuracy", "sem"]].sort_values(["accuracy", "sem"], ascending=[False, True])
            idx_bestscore = df_stats.index[0]
            best_score = df_stats.loc[idx_bestscore,"accuracy"]
            sem_at_best_score = df_stats.loc[idx_bestscore, "sem"]

            # Lines
            ax.axhline(best_score, linestyle=":", color="black", label= "%s: %0.3f ± %0.3f"%(score_metric, best_score, sem_at_best_score))
            ax.axvline(idx_bestscore, linestyle=":", color="black", label=f"n= {idx_bestscore} {self.attr_type}s")

            if baseline_score is not None:
                ax.axhline(baseline_score, color="maroon", linestyle=":", label=f"Baseline: {baseline_score}")

            # Labels
            if ylabel is None:
                ylabel = score_metric
            ax.set_xlabel(f"Number of {self.attr_type}s included")
            ax.set_ylabel(ylabel, fontsize=12)
            ax.set_xlabel(ax.get_xlabel(), fontsize=12)
            ax.legend(**_legend_kws)
            if title is not None:
                ax.set_title(title, _title_kws)
            if ylim:
                ax.set_ylim(ylim)
            pad = 0.1
            ax.set_xlim((scores.index.min()-pad, scores.index.max()+pad))
            return fig, ax


    def plot_weights(self,  ax=None, style="seaborn-white",  title=None, title_kws=dict(),  axis_kws=dict(), tick_kws=dict(), figsize=(21,5), palette="hls", *args):
        """
        Plot the weights
        """
        _title_kws = {"fontsize":15, "fontweight":"bold"}
        _title_kws.update(title_kws)

        _axis_kws = {"fontsize":15}
        _axis_kws.update(axis_kws)

        _tick_kws = {"fontsize":12}
        _tick_kws.update(tick_kws)
        with plt.style.context(style):
            if ax is None:
                fig, ax = plt.subplots(figsize=figsize)
            else:
                fig = plt.gcf()

            df_w = self.extract_weights(mode=1, *args)
            color = sns.color_palette(palette=palette, n_colors=df_w.shape[1])
            df_w.plot(color=color, ax=ax)
            ax.legend_.remove()
            ax.set_xlabel("Threshold", **_axis_kws)
            ax.set_ylabel("$W$", **_axis_kws)
            ax.set_xticklabels(ax.get_xticks(), **_tick_kws)
            ax.set_yticklabels(ax.get_yticks(), **_tick_kws)
            ax.set_xlim((df_w.index.min(), df_w.index.max()))
            if title is not None:
                ax.set_title(title, _title_kws)
            return fig, ax

    # Writing serialized object
    def save_model(self, path, compression="infer"):
        """
        Extensions:
        pickle ==> .pkl
        dill ==> .dill
        gzipped-pickle ==> .pgz
        bzip2-pickle ==> .pbz2
        """
        write_object(self, path)



# ==============================================================================
# Imports
# ==============================================================================
# Built-ins
import os, sys
from collections import defaultdict, OrderedDict
import pandas as pd
import numpy as np
from statsmodels.sandbox.stats.multicomp import multipletests
# Soothsayer
from ..utils import is_dict, assert_acceptable_arguments

__all__ = ["p_adjust", "statistically_significant_symbols", "biweight_midcorrelation"]

# Adjust p-values
def p_adjust(p_values:pd.Series, method="fdr", name=None, **kwargs):
    """
    Multiple test correction to adjust p-values

    fdr -> fdr_bh # https://www.statsmodels.org/stable/generated/statsmodels.stats.multitest.multipletests.html

    `bonferroni` : one-step correction
    `sidak` : one-step correction
    `holm-sidak` : step down method using Sidak adjustments
    `holm` : step-down method using Bonferroni adjustments
    `simes-hochberg` : step-up method  (independent)
    `hommel` : closed method based on Simes tests (non-negative)
    `fdr_bh` : Benjamini/Hochberg  (non-negative)
    `fdr_by` : Benjamini/Yekutieli (negative)
    `fdr_tsbh` : two stage fdr correction (non-negative)
    `fdr_tsbky` : two stage fdr correction (non-negative)

    """
    # Check type of input data
    index = None
    if is_dict(p_values):
        p_values = pd.Series(p_values)
    if isinstance(p_values, pd.Series):
        index = p_values.index
        p_values = p_values.values
    if method == "fdr":
        method = "fdr_bh"

    if index is not None:
        mask_null = np.isnan(p_values)
        if np.any(mask_null):
            scaffold = np.ones_like(p_values)*np.nan
            scaffold[np.logical_not(mask_null)] = multipletests(p_values[np.logical_not(mask_null)], method=method, **kwargs)[1]
            return pd.Series(scaffold, index=index)
        else:
            adjusted = multipletests(p_values, method=method, **kwargs)[1]
            return pd.Series(adjusted, index=index, name=name)
    else:
        num_nan = np.isnan(p_values).sum()
        assert  num_nan == 0, "Please remove the {} missing values".format(num_nan)
        return multipletests(p_values, method=method, **kwargs)[1]

# Statistically significant symbols
def statistically_significant_symbols(p:pd.Series, not_significant="ns", return_pvalues=False):
    """
    # Lexicon
    # ns
    # P > 0.05
    # *
    # P ≤ 0.05
    # **
    # P ≤ 0.01
    # ***
    # P ≤ 0.001
    # ****
    #  P ≤ 0.0001 (For the last two choices only)

    Future:  Make this customizable
    """
    if not hasattr(p, "__iter__"):
        if p > 0.05:
            return not_significant
        symbol = ""
        if p <= 0.05:
            symbol += "*"
        if p <= 0.01:
            symbol += "*"
        if p <= 0.001:
            symbol += "*"
        if p <= 0.0001:
            symbol += "*"
        return symbol
    else:
        symbols =  pd.Series(p).map(lambda x:statistically_significant_symbols(x, not_significant=not_significant))
        if return_pvalues:
            return pd.concat([symbols.to_frame("symbol"), p.to_frame("p_value")], axis=1)
        else:
            return symbols

# Biweight midcorrelation
def biweight_midcorrelation(a,b, check_index_order=True, use_numba=False, verbose=False):
    """
    a,b: {np.array, pd.Series}

    Code adapted from the following sources:
        * https://stackoverflow.com/questions/61090539/how-can-i-use-broadcasting-with-numpy-to-speed-up-this-correlation-calculation/61219867#61219867
        * https://github.com/olgabot/pandas/blob/e8caf4c09e1a505eb3c88b475bc44d9389956585/pandas/core/nanops.py

    Special thanks to the following people:
        * @norok2 (https://stackoverflow.com/users/5218354/norok2) for optimization (vectorization and numba)
        * @olgabot (https://github.com/olgabot) for NumPy implementation

    Benchmarking:
        * iris_features (4)
            * numba: 321 ms ± 15.3 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)
            * numpy: 478 µs ± 8.37 µs per loop (mean ± std. dev. of 7 runs, 1000 loops each)
        * iris_samples: (150)
            * numba: 312 ms ± 8.54 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)
            * numpy: 438 µs ± 5.57 µs per loop (mean ± std. dev. of 7 runs, 1000 loops each)

    Future:
        * Handle missing values

    """
    # Data
    result = None
    labels = None
    assert type(a) is type(b), "a & b must both either be np.array or pd.Series"
    if isinstance(a, pd.Series):
        if check_index_order:
            assert np.all(a.index == b.index), "a.index and b.index must be be the same ordering"
        labels = a.index
        a = a.values
        b = b.values

    # Base computation
    def _base_computation(a,b):
        n = a.size
        a = a - np.median(a)
        b = b - np.median(b)
        v_a = 1 - (a / (9 * np.median(np.abs(a)))) ** 2
        v_b = 1 - (b / (9 * np.median(np.abs(b)))) ** 2
        a = (v_a > 0) * a * v_a ** 2
        b = (v_b > 0) * b * v_b ** 2
        return n, a, b

    # Check if numba is available
    assert_acceptable_arguments(use_numba, {True, False, "infer"})
    if use_numba == "infer":
        if "numba" in sys.modules:
            use_numba = True
        else:
            use_numba = False
        if verbose:
            print("Numba is available:", use_numba, file=sys.stderr)

    # Compute using numba
    if use_numba:
        assert "numba" in sys.modules
        from numba import jit

        @jit
        def _biweight_midcorrelation_numba(a, b):
            n, a, b = _base_computation(a,b)
            s_ab = s_aa = s_bb = 0
            for i in range(n):
                s_ab += a[i] * b[i]
                s_aa += a[i] * a[i]
                s_bb += b[i] * b[i]
            return s_ab / np.sqrt(s_aa) / np.sqrt(s_bb)

        result = _biweight_midcorrelation_numba(a,b)

    # Compute using numpy
    else:
        def _biweight_midcorrelation_numpy(a, b):
            n, a, b = _base_computation(a,b)
            return np.sum(a * b) / (np.sqrt(np.sum(a ** 2)) * np.sqrt(np.sum(b ** 2)))
        result = _biweight_midcorrelation_numpy(a,b)

    # Add labels
    if labels is not None:
        result = pd.Series(result, index=labels)

    return result



# ==============================================================================
# Modules
# ==============================================================================
# Built-ins
import os, sys, time, datetime, copy, pathlib
from collections import OrderedDict, defaultdict, Mapping
from io import StringIO, BytesIO
# from ast import literal_eval

# Compression & Serialization
import pickle, gzip, bz2, zipfile

# PyData
import pandas as pd

# Soothsayer
from soothsayer.utils import format_path, to_precision

# ======
# Future
# ======
# (1) Add ability for `path` arguments to be `pathlib.Path`
# (2) Add ability to process ~ in path
# ======
# Tables
# ======
# Read dataframe
def read_dataframe(path:str, sep="infer", index_col=0, header=0, compression="infer", pickled="infer", func_index=None, func_columns=None, evaluate_columns=None, engine="c", verbose=False, excel="infer", infer_series=False, sheet_name=None,  **args):

    start_time = time.time()
    path = format_path(path, str)
    dir , ext = os.path.splitext(path)
    ext = ext.lower()

    if excel == "infer":
        if ext in {".xlsx", ".xls"}:
            excel = True
        else:
            excel = False

    if excel:
        if "sheetname" in args:
            sheet_name = args.pop("sheetname")
            print("DEPRECATED: Use `sheet_name` instead of `sheetname`", file=sys.stderr)
        df = pd.read_excel(path, sheet_name=sheet_name, index_col=index_col, header=header, **args)

    else:
        # Seperator
        if any(list(map(lambda x:ext.endswith(x),[".csv", "csv.gz", "csv.zip"]))):
            sep = ","
        else:
            sep = "\t"

        # Serialization
        if pickled == "infer":
            if ext in {".pkl", ".pgz", ".pbz2"}:
                pickled = True
            else:
                pickled = False
        # Compression
        if compression == "infer":
            if pickled:
                if ext == ".pkl":
                    compression = None
                if ext == ".pgz":
                    compression = "gzip"
                if ext == ".pbz2":
                    compression = "bz2"
            else:
                if ext == ".gz":
                    compression = "gzip"
                if ext == ".bz2":
                    compression = "bz2"
                if ext == ".zip":
                    compression = "zip"


        if pickled:
            df = pd.read_pickle(path, compression=compression)
        else:
            df = pd.read_csv(path, sep=sep, index_col=index_col, header=header, compression=compression, engine=engine, **args)

    condition_A = any([(excel == False), (sheet_name is not None)])
    condition_B = all([(excel == True), (sheet_name is None)])

    if condition_A:
        # Map indices
        if func_index is not None:
            df.index = df.index.map(func_index)
        if func_columns is not None:
            df.columns = df.columns.map(func_columns)

        if evaluate_columns is not None:
            for field_column in evaluate_columns:
                try:
                    df[field_column] = df[field_column].map(eval)
                except ValueError:
                    if verbose:
                        print(f"Could not use `eval` on column=`{field_column}`", file=sys.stderr)
        if verbose:
            print(f"{path.split('/')[-1]} | Dimensions: {df.shape} | Time: {to_precision(time.time() - start_time)} seconds", file=sys.stderr)
    if condition_B:
        if verbose:
            print(f"{path.split('/')[-1]} | Sheets: {len(df)} | Time: {to_precision(time.time() - start_time)} seconds", file=sys.stderr)
    if infer_series:
        if df.shape[1] == 1:
            df = df.iloc[:,0]
    return df

# Write dataframe
def write_dataframe(data, path:str, sep="\t", compression="infer", pickled="infer",  excel="infer", create_directory_if_necessary=False, **args):
    start_time = time.time()
    path = format_path(path, str)
    _ , ext = os.path.splitext(path)
    dir, filename = os.path.split(path)
    if not os.path.exists(dir):
        dir = str(pathlib.Path(dir).absolute())
        os.makedirs(dir, exist_ok=True)

    # Excel
    if excel == "infer":
        if ext in {".xlsx", ".xls"}:
            excel = True
        else:
            excel = False

    if excel:
        if not isinstance(data, Mapping):
            data = {"Sheet1":data}
        writer = pd.ExcelWriter(path)
        for sheet_name, df in data.items():
            df.to_excel(writer, sheet_name=str(sheet_name))
        writer.save()
    else:
        # Serialization
        if pickled == "infer":
            if ext in {".pkl", ".pgz", ".pbz2"}:
                pickled = True
            else:
                pickled = False
        # Compression
        if compression == "infer":
            if pickled:
                if ext == ".pkl":
                    compression = None
                if ext == ".pgz":
                    compression = "gzip"
                if ext == ".pbz2":
                    compression = "bz2"
            else:
                compression = None
                if ext == ".gz":
                    compression = "gzip"
                if ext == ".bz2":
                    compression = "bz2"

        if pickled:
            data.to_pickle(path, compression=compression, **args)
        else:
            data.to_csv(path, sep=sep, compression=compression, **args)

# Read results from feature selection
def read_multiple_synopsis(dir:str, from_algorithm:str="clairvoyance"):
    """
    Assumes the following directory structure:
    __________________________________
    | project_directory
    | | synopsis_dir_for_all_submodels
    | | | {name}_synopsis
    | | | | {name}__synopsis.tsv
    ----------------------------------
    """
    accepted_args = ["clairvoyance"]
    def _read_clairvoyance(dir:str):
        # Get synopsis
        _data = list()
        for file in os.scandir(dir):
            if file.is_dir():
                submodel = file.name.split("_")[1]
                for file2 in os.scandir(file):
                    if file2.name.endswith("__synopsis.tsv"):
                        _df = read_dataframe(file2)
                        _df.insert(loc=0, column="submodel", value=submodel)
                        _data.append(_df)
        df_synopsis = pd.concat(_data, axis=0).reset_index()

        # Evaluate fields
        fields_to_evaluate = ["hyperparameters"] + [*filter(lambda field:field.endswith("set"), df_synopsis.columns)]
        for field in fields_to_evaluate:
            df_synopsis[field] = df_synopsis[field].map(eval)
        return df_synopsis

    assert from_algorithm in accepted_args, f"{from_algorithm} is not in {accepted_args}"
    dir = format_path(dir, str)
    if from_algorithm == "clairvoyance":
        return _read_clairvoyance(dir)



# ==============================================================================
# Imports
# ==============================================================================
# Built-ins
import os, sys, time
from collections import namedtuple, OrderedDict
import pandas as pd
import numpy as np
import xarray as xr

# Soothsayer
from ..r_wrappers import *
from soothsayer.utils import assert_acceptable_arguments, check_packages

# ==============================================================================
# R Imports
# ==============================================================================
if "rpy2" in sys.modules:
    from rpy2 import robjects as ro
    from rpy2 import rinterface as ri

    from rpy2.robjects.packages import importr
    try:
        from rpy2.rinterface import RRuntimeError
    except ImportError:
        from rpy2.rinterface_lib.embedded import RRuntimeError
    from rpy2.robjects import pandas2ri
    # pandas2ri.activate()
    R = ro.r
    NULL = ri.NULL
    #rinterface.set_writeconsole_regular(None)

# R Packages
# propr = R_package_retrieve("propr")

# ==============================================================================
# Exports
# ==============================================================================
__all__ = ["propr"]

# ==============================================================================
# Functions
# ==============================================================================
# Proportionality
@check_packages(["propr"], language="r", import_into_backend=False)
def proportionality(X:pd.DataFrame, metric="rho", transformation="clr", symmetrize=False,  permutations=100,  store_counts=False, store_transformed_components=True, name=None, into=OrderedDict, **kwargs):
    """
    https://github.com/tpq/propr
    https://rdrr.io/cran/propr/man/propr.html
    https://journals.plos.org/ploscompbiol/article?id=10.1371/journal.pcbi.1004075
    
    X
    A pd.DataFrame. Stores the original "count matrix" input.

    alpha
    A double. Stores the alpha value used for transformation.

    metric
    A character string. The metric used to calculate proportionality.

    ivar
    A vector. The reference used to calculate proportionality.

    logratio
    A data.frame. Stores the transformed "count matrix".

    matrix
    A matrix. Stores the proportionality matrix.

    pairs
    A vector. Indexes the proportional pairs of interest.

    results
    A data.frame. Stores the pairwise propr measurements.

    permutes
    A list. Stores the shuffled transformed "count matrix" instances, used to reproduce permutations of propr.

    fdr
    A data.frame. Stores the FDR cutoffs for propr.
    """
    # Imports
    propr = R_package_retrieve("propr")

    # Assert "into" types
    assert_acceptable_arguments([into], {dict, OrderedDict, namedtuple})
    # Run propr
    obj_propr = propr.propr(pandas_to_rpy2(X), metric=metric, ivar=transformation,  p=permutations, symmetrize=symmetrize,**kwargs )
#     if update_cutoffs:
#         if n_jobs == -1:
#             n_jobs = multiprocessing.cpu_count()
#         # Cutoffs at which to estimate FDR
#         propr.updateCutoffs(obj_propr, cutoff = ro.FloatVector(cutoff), ncores = n_jobs) 
    
    # Free up some space
    results = dict(obj_propr.slots)
    del obj_propr
    
    # Get useful results
    output = OrderedDict()
    
    # Counts
    if store_counts:
        df = rpy2_to_pandas(results['counts'])
        df.index = X.index
        output["counts"] = df
        
    # Proportinoality
    output["proportionality"]  = pd.DataFrame(np.asarray(results["matrix"]), index=X.columns, columns=X.columns)
    
    # Transformed components
    output["transformation"] = results["ivar"]
    if store_transformed_components:
        output["transformed_components"] = results["logratio"]
        
    # Permutations
    data = np.asarray(list(map(np.asarray, results["permutes"])))
    output["permutations"] = xr.DataArray(data=data, dims=["Permutations", "Samples", "Components"], coords={"Permutations":range(permutations), "Samples":X.index, "Components":X.columns})
    
    # FDR 
    output["fdr"] = rpy2_to_pandas(results["fdr"])
    # Output
    if into is namedtuple:
        if name is None:
            name = "Proportionality"
        Proportionality = namedtuple(name, field_names=output.keys())
        return Proportionality(**output)
    else:
        return into(output)


# ==============================================================================
# Modules
# ==============================================================================
# Built-ins
import os, sys, time, pathlib, warnings, requests
from collections import *
from io import StringIO, BytesIO
from ast import literal_eval
from tqdm import tqdm
import xml.etree.ElementTree as ET

# Compression & Serialization
import pickle, gzip, bz2, zipfile

# PyData
import pandas as pd
import numpy as np

# Biology
from Bio import SeqIO

# Soothsayer
from soothsayer.utils import format_path, infer_compression, is_path_like, is_query_class, assert_acceptable_arguments
from ..text import read_textfile, get_file_object

# ======
# Future
# ======
# (1) Add ability for `path` arguments to be `pathlib.Path`
# (2) Add ability to process ~ in path

# ==============
# Bioinformatics
# ==============
# Clustalo Distance Matrix
def read_clustalo_distmat(path:str):
    path = format_path(path)
    with open(path, "r") as f:
        for i, line in enumerate(f):
            line = line.strip()
            conditions = [
                line.startswith("#"),
            ]
            try:
                int(line)
                conditions.append(True)
            except ValueError:
                pass
            if not any(conditions):
                break
    df_tmp = pd.read_csv(path, sep="\t", index_col=0, header=None, skiprows=i)
    tmp = df_tmp.index.map(lambda line:[x for x in line.split(" ") if len(x) > 0])
    if ":" in tmp[0][0]:
        tmp = tmp.map(lambda x:x[1:])
    data = OrderedDict()
    try:
        for line in tmp:
            data[line[0]] = pd.Series([*map(float,line[1:])], name=line[0])
        df_dism = pd.DataFrame(data)
        df_dism.index = df_dism.columns
        return df_dism
    except ValueError:
        print("Could not accurately parse clustal distance matrix.  Returning raw data instead.", file=sys.stderr)
        return tmp

# VizBin corrdinates
def read_vizbin_coords(xy_path:str, fasta_path:str=None, xlabel="x", ylabel="y", write_coords=True):
    """
    VizBin: http://claczny.github.io/VizBin/
    xy_path: VizBin coordinates file path
    fasta_path: Input fasta file path

    Usage:
    xy_path = "/var/folders/6z/5vbtz_gmkr76ftgc3149dvtr0003c0/T/map1723096415078571924/points.txt"
    fasta_path = "./assembly.fa")
    df_xy = read_vizbin_coords(xy_path, fasta_path)
    """
    xy_path = format_path(xy_path)

    df_xy = pd.read_csv(xy_path, sep=",", header=None).astype(float)
    path_write = None
    if fasta_path is not None:
        fasta_path = format_path(fasta_path)
        df_xy.index = [*map(lambda record:str(record.id), SeqIO.parse(fasta_path, "fasta"))]
        if write_coords == True:
            path_write = f"{fasta_path}.xy"
    df_xy.columns = [xlabel,ylabel]
    if path_write is not None:
        df_xy.to_csv(path_write, sep="\t")
    return df_xy

# CheckM QA
def read_checkm_qa(path:str, format=8, verbose=True, rename_columns=True, func_orf_to_contig=lambda id_orf:"_".join(id_orf.split("_")[:-1])):
    """
    Tested for v1.0.7
    Read CheckM qa output into pd.DataFrame
    accepted_formats = {2,8}
    """
    accepted_formats = {2,8}
    assert format in accepted_formats, f"Currently only supporting formats: `{accepted_formats}`"
    path = format_path(path)

    # QA: Format = 2
    def _read_format_qa_2(path, rename_columns, verbose):
        d_rename = {
            "Bin Id":"id_bin",
            "Marker lineage":"marker_lineage",
            "# genomes":"number_of_genomes",
            "# markers":"number_of_markers",
            "# marker sets":"number_of_marker-sets",
            "Completeness":"completeness",
            "Contamination":"contamination",
            "Strain heterogeneity":"strain_heterogeneity",
            "Genome size (bp)":"genome_size[bp]",
            "# ambiguous bases":"number_of_ambiguous-bases",
            "# scaffolds":"number_of_scaffolds",
            "# contigs":"number_of_contigs",
            "N50 (scaffolds)":"N50[scaffolds]",
            "N50 (contigs)":"N50[contigs]",
            "Mean scaffold length (bp)":"Mean_scaffold_length[bp]",
            "Mean contig length (bp)":"Mean_contig_length[bp]",
            "Longest scaffold (bp)":"Longest_scaffold[bp]",
            "Longest contig (bp)":"Longest_contig[bp]",
            "GC std (scaffolds > 1kbp)":"GC_std[scaffolds>1kbp]",
            "Coding density":"coding_density",
            "Translation table":"trans_table",
            "# predicted genes":"number_of_predicted-genes",
        }
        tab_table = False
        for i, line in read_textfile(path):
            if i == 0:
                if not line.startswith("-"):
                    tab_table = True
                    break
            if i == 1:
                columns = pd.Index([*map(lambda x: x.strip(), filter(bool, line.split("  ")))])
                break
        if tab_table:
            df_table = pd.read_csv(path, sep="\t", index_col=None)
            columns = df_table.columns
        else:
            table = pd.read_csv(path, sep="\t", skiprows=[0,1,2],skipfooter=1, header=None).values.ravel()
            df_table = pd.DataFrame([*map(lambda line: [*filter(bool, line.split(" "))], table)])
            df_table.columns = columns
        if verbose:
            print("Total bins:", df_table.shape[0], file=sys.stderr)
        if rename_columns:
            df_table.columns = columns.map(lambda c: d_rename[c] if c in d_rename else c)
            return df_table.set_index("id_bin", drop=True).sort_values(["completeness", "contamination"], ascending=[False, True])
        else:
            return df_table.set_index("Bin Id", drop=True).sort_values(["Completeness", "Contamination"], ascending=[False, True])
    # QA: Format = 8
    def _read_format_qa_8(path, verbose):
        table = list()
        for i,line in read_textfile(path):
            if i > 0:
                data = line.replace("\n","").split("\t")
                # Get bin ID
                ids = data[:2]
                # Get HMM IDs
                hmm_ids = set([x.split(".")[0].split(",")[0] for x in data[2:]  ])
                # Concat positions
                if i == 0:
                    hmm_pos = data[2:]
                else:
                    hmm_pos = set([*map(lambda organized_hit: (organized_hit[0], int(organized_hit[1]), int(organized_hit[2])), map(lambda hit: tuple(hit.split(",")), data[2:]))])
                # Create table
                line_out = ids + [hmm_ids] + [hmm_pos]
                table.append(line_out)
        # Build table
        df_table = pd.DataFrame(table)
        df_table.columns = ["id_bin", "id_orf", "markers", "marker_positions"]

        # Split ORFs
        table = list()
        for i, row in df_table.iterrows():
            if "&&" in row["id_orf"]:
                for id_orf in row["id_orf"].split("&&"):
                    sub_row = row.copy()
                    sub_row["id_orf"] = id_orf
                    table.append(sub_row)
            else:
                table.append(row)
        df_table = pd.DataFrame(table)

        if func_orf_to_contig is not None:
            df_table.insert(loc=1, column="id_contig", value=df_table["id_orf"].map(func_orf_to_contig).tolist())
        if verbose:
            print("Total markers:",len(set().union(*df_table["markers"])), sep="\t", file=sys.stderr)
        return df_table

    # Controller
    if format == 2:
        return _read_format_qa_2(path, rename_columns, verbose)
    if format == 8:
        return _read_format_qa_8(path, verbose)

# Read Fasta File
def read_fasta(path:str, description:bool=True, case="upper", func_header=None, into=pd.Series, compression="infer", name=None, verbose=True):
    """
    Reads in a single fasta file or a directory of fasta files into a dictionary.
    """
    # Get path
    path = format_path(path)

    # Assign pathname as name if there isn't one
    if name is None:
        name = path

    # Open file object
    f = get_file_object(path, mode="read", compression=compression, safe_mode=False, verbose=False)

    # Read in fasta
    d_id_seq = OrderedDict()

    if verbose:
        seq_records = tqdm(SeqIO.FastaIO.SimpleFastaParser(f), f"Reading sequence file: {path}")
    else:
        seq_records = SeqIO.FastaIO.SimpleFastaParser(f)

    # Verbose but faster
    if case == "lower":
        for header, seq in seq_records:
            seq = seq.lower()
            if not description:
                header = header.split(" ")[0]
            d_id_seq[header] = seq
    if case == "upper":
        for header, seq in seq_records:
            seq = seq.upper()
            if not description:
                header = header.split(" ")[0]
            d_id_seq[header] = seq
    if case is None:
        for header, seq in seq_records:
            if not description:
                header = header.split(" ")[0]
            d_id_seq[header] = seq

    # Close File
    f.close()

    # Transform header
    if func_header is not None:
        d_id_seq = OrderedDict( [(func_header(id),seq) for id, seq in d_id_seq.items()])
    sequences = into(d_id_seq)
    if hasattr(sequences, "name"):
        sequences.name = name
    return sequences


# Writing sequence files
def write_fasta(sequences, path:str, compression="infer"):
    """
    Sequence stats:
    count    29999.000000
    mean       310.621754
    std       1339.422833
    min         56.000000
    25%         75.000000
    50%        111.000000
    75%        219.000000
    max      54446.000000

    Benchmarks:
    No compression: 47.2 ms ± 616 µs per loop (mean ± std. dev. of 7 runs, 10 loops each)
    Gzip: 9.85 s ± 261 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)
    Bzip2: 864 ms ± 16.1 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)
    """
    # path = format_path(path)
    # if compression == "infer":
    #     compression = infer_compression(path)
    if is_query_class(path, ["stdout", "stderr", "streamwrapper"]):
        path.writelines(f">{id}\n{seq}\n" for id, seq in sequences.items())
    else:
        with get_file_object(path, mode="write", compression=compression, safe_mode=False, verbose=False) as f:
            f.writelines(f">{id}\n{seq}\n" for id, seq in sequences.items())

# Read blast output
def read_blast(path:str, length_query=None, length_subject=None, sort_by="bitscore"):
    """
    if 12: ['qseqid', 'sseqid', 'pident', 'length', 'mismatch', 'gapopen', 'qstart', 'qend', 'sstart', 'send', 'evalue', 'bitscore']
    if 15 assumes: -outfmt '6 std qlen slen stitle': ["std", "qlen", "slen", "stitle"]
    ####################################################
    Column    NCBI name    Description
    1    qseqid    Query Seq-id (ID of your sequence)
    2    sseqid    Subject Seq-id (ID of the database hit)
    3    pident    Percentage of identical matches
    4    length    Alignment length
    5    mismatch    Number of mismatches
    6    gapopen    Number of gap openings
    7    qstart    Start of alignment in query
    8    qend    End of alignment in query
    9    sstart    Start of alignment in subject (database hit)
    10    send    End of alignment in subject (database hit)
    11    evalue    Expectation value (E-value)
    12    bitscore    Bit score
    13    sallseqid    All subject Seq-id(s), separated by a ';'
    14    score    Raw score
    15    nident    Number of identical matches
    16    positive    Number of positive-scoring matches
    17    gaps    Total number of gaps
    18    ppos    Percentage of positive-scoring matches
    19    qframe    Query frame
    20    sframe    Subject frame
    21    qseq    Aligned part of query sequence
    22    sseq    Aligned part of subject sequence
    23    qlen    Query sequence length
    24    slen    Subject sequence length
    25    salltitles    All subject title(s), separated by a '<>'

    Example inputs:
        * blat -prot Yeast/Saccharomyces_cerevisiae.R64-1-1.pep.all.processed.fa  Phaeodactylum_tricornutum.ASM15095v2.pep.all.processed.fa -out=blast8 yeast-pt.blast8
        * diamond blastp -d ../../../../reference_data/references/gracilibacteria/reference_proteins.nmnd -q ./prodigal_output/orfs.faa -f 6 -o ./diamond_output/output.blast6

    """
    path = format_path(path)
    idx_default_fields = ['qseqid', 'sseqid', 'pident', 'length', 'mismatch', 'gapopen', 'qstart', 'qend', 'sstart', 'send', 'evalue', 'bitscore']
    # ["query_id", "subject_id", "identity", "alignment_length", "mismatches", "gap_openings", "query_start", "query_end", "subject_start", "subject_end", "e-value", "bit_score"]
    df_blast = pd.read_csv(path, header=None, sep="\t")
    if df_blast.shape[1] == 12:
        df_blast.columns = idx_default_fields
    if df_blast.shape[1] == 15:
        df_blast.columns = idx_default_fields + ["qlen", "slen", "stitle"]
    # Length of query
    if length_query is not None:
        if is_path_like(length_query):
            length_query = read_fasta(length_query, description=False, verbose=False)
        if isinstance(length_query[0], str):
            length_query = length_query.map(len)
        df_blast["qlength"] = df_blast["qseqid"].map(lambda id: length_query[id])
        df_blast["qratio"] = (df_blast["qend"] - df_blast["qstart"])/df_blast["qlength"]
    # Length of subject
    if length_subject is not None:
        if is_path_like(length_subject):
            length_subject = read_fasta(length_subject, description=False, verbose=False)
        if isinstance(length_subject[0], str):
            length_subject = length_subject.map(len)
        df_blast["slength"] = df_blast["sseqid"].map(lambda id: length_subject[id])
        df_blast["sratio"] = (df_blast["send"] - df_blast["sstart"])/df_blast["slength"]
    if sort_by is not None:
        df_blast = df_blast.sort_values(by=sort_by, ascending=False).reset_index(drop=True)

    return df_blast
# Helper function for reading gtf and gff3
def _read_gtf_gff_base(path, compression, record_type, verbose):
    # Read the gff3 file
    with get_file_object(path, mode="read", compression=compression, safe_mode=False, verbose=False) as f:
        if verbose:
            iterable_lines = tqdm(f.readlines(), "Processing lines")
        else:
            iterable_lines = f.readlines()
        data= list()
        if record_type is None:
            for line in iterable_lines:
                if not line.startswith("#"):
                    line = line.strip("\n")
                    if bool(line):
                        base_fields = line.split("\t")
                        data.append(base_fields)
        else:
            for line in iterable_lines:
                if not line.startswith("#"):
                    if f"{record_type}_id" in line:
                        line = line.strip("\n")
                        base_fields = line.split("\t")
                        data.append(base_fields)
    # Generate table
    df_base = pd.DataFrame(data)
    df_base.columns = ["seq_record", "source", "seq_type", "pos_start", "pos_end", ".1", "sense", ".2", "data_fields"]
    return df_base

def read_gff3(path:str, compression="infer", record_type=None, verbose = True, reset_index=False, name=True):
    def f(x):
        fields = x.split(";")
        data = dict()
        for item in fields:
            k, v = item.split("=")
            data[k] = v
        return data
    path = format_path(path)
    if verbose:
        print("Reading gff3 file:",path,sep="\t", file=sys.stderr)
    accepted_recordtypes = {"exon", "gene", "transcript", "protein", None}
    assert record_type in accepted_recordtypes, "Unrecognized record_type.  Please choose from the following: {}".format(accepted_recordtypes)

    # Read the gff3 file
    df_base = _read_gtf_gff_base(path, compression, record_type, verbose)

    try:
        df_fields = pd.DataFrame(df_base["data_fields"].map(f).to_dict()).T
        df_gff3 = pd.concat([df_base[["seq_record", "source", "seq_type", "pos_start", "pos_end", ".1", "sense", ".2"]], df_fields], axis=1)
         # Contains identifier
        if "ID" in df_gff3.columns:
            df_gff3["seq_id"] = df_gff3["ID"].map(lambda x: "-".join(x.split("-")[1:]) if not pd.isnull(x) else x)
        if reset_index:
            df_gff3 = df_gff3.reset_index(drop=True)
    except IndexError:
        warnings.warn("Could not successfully parse file: {}".format(path))
        df_gff3 = df_base

    # Path index
    if name is not None:
        if name == True:
            name = path
        df_gff3.index.name = name

    return df_gff3

# GTF
def read_gtf(path:str, compression="infer", record_type=None, verbose = True, reset_index=False, name=True):
    path = format_path(path)
    if verbose:
        print(f"Reading gtf file:",path,sep="\t", file=sys.stderr)
    accepted_recordtypes = {"exon", "gene", "transcript", "protein", None}
    assert record_type in accepted_recordtypes, f"Unrecognized record_type.  Please choose from the following: {accepted_recordtypes}"

    # Read the gtf file
    df_base = _read_gtf_gff_base(path, compression, record_type, verbose)

    # Splitting fields
    iterable_fields =  df_base.iloc[:,-1].iteritems()

    # Parse fields
    dataframe_build = dict()
    for i, gtf_data in iterable_fields:
        gtf_data = gtf_data.replace("''","").replace('"',"")
        fields = filter(bool, map(lambda field:field.strip(), gtf_data.split(";")))
        fields = map(lambda field: field.split(" "), fields)
        dataframe_build[i] = dict(fields)
    df_fields = pd.DataFrame(dataframe_build).T
    df_gtf = pd.concat([df_base.iloc[:,:7], df_fields], axis=1)

    # Reset index
    if reset_index:
        df_gtf = df_gtf.reset_index(drop=True)

    # Path index
    if name is not None:
        if name == True:
            name = path
        df_gtf.index.name = name
    return df_gtf

# NCBI XML
def read_ncbi_xml(path:str, index_name=None):
    # Format path
    path = format_path(path)
    # Parse the XML tree
    tree = ET.parse(path)
    root = tree.getroot()
    # Extract the attributes
    data = defaultdict(dict)
    for record in tqdm(root.getchildren(), "Reading NCBI XML: {}".format(path)):
        id_record = record.attrib["accession"]
        for attribute in record.findall("Attributes/*"):
            data[id_record][attribute.attrib["attribute_name"]] = attribute.text
    # Create pd.DataFrame
    df = pd.DataFrame(data).T
    df.index.name = index_name
    return df

# Read EMBL-EBI sample metadata
def read_ebi_sample_metadata(query, base_url="https://www.ebi.ac.uk/metagenomics/api/v1/samples/{}/metadata", mode="infer"):
    """
    query:
        id_sample: "ERS488919"
        url: "https://www.ebi.ac.uk/metagenomics/api/v1/samples/ERS488919/metadata
        file: "/path/to/ERS488919.json"

    JSON file can be retrieved via:
        curl -X GET "https://www.ebi.ac.uk/metagenomics/api/v1/samples/ERS488919/metadata" -H "accept: application/json" > ERS488919.json
    """
    # Acceptable arguments for mode
    assert_acceptable_arguments(mode, {"infer", "cloud", "local"})

    # Functions
    def _get_url(query, base_url):
        # Is a URL provided?
        if any(map(lambda x: x in query, ["http", "www"])):
            url = query
        # Is a sample identifier provided
        else:
            url = base_url.format(query)
        return url

    # Infer mode
    url = None
    response = None
    if mode == "infer":
        try:
            url = _get_url(query, base_url)
            response = requests.get(url)
            mode = "cloud"
        except requests.exceptions.MissingSchema:
            mode = "local"

    # Cloud
    if mode == "cloud":
        if response is None:
            # Get cloud data
            url = _get_url(query, base_url)
            response = requests.get(url)
        data = response.json()["data"]
    if mode == "local":
        with open(query, "r") as f:
            data = json.loads(f.read())["data"]
    # Determine sample identifier
    id_sample = data[0]["relationships"]["sample"]["data"]["id"]

    return pd.Series(
            data=dict(map(lambda item: (item["attributes"]["key"], item["attributes"]["value"]), data)),
            name=id_sample,
    )
